from argparse import ArgumentParser, ArgumentTypeError, HelpFormatter, Namespace
from numbers import Number
import operator
from pathlib import Path
import re
import textwrap
from typing import Callable, Sequence, List, Type, TypeVar, Generic, Union, Optional
from urllib.parse import urlparse

from xphyle.paths import (
    STDOUT,
    STDERR,
    check_path,
    as_path,
    resolve_path,
    check_readable_file,
    check_writable_file,
)
from xphyle.types import PathLike, PathType, Permission

from atropos.utils import Magnitude


class AtroposArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._groups = {}

    def add_group(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        mutex: bool = False,
        required: bool = False,
    ):
        """
        Add a group to the parser. The group will be stored under `name` and can
        later be retrieved via `get_group`.
        """
        if name in self._groups:
            raise ValueError(f"Group already exists: {name}")

        if mutex:
            group = self.add_mutually_exclusive_group(required=required)
        else:
            group = self.add_argument_group(title or name, description)
        self._groups[name] = group
        return group

    def get_group(self, name: str):
        """
        If a group has already been created with `name`, return the group, otherwise
        create a new group with that name.
        """
        if name in self._groups:
            return self._groups[name]
        else:
            return self.add_group(name)


class ParagraphHelpFormatter(HelpFormatter):
    """
    HelpFormatter that wraps text in paragraphs.
    """

    def _fill_text(self, text, width, indent):
        text = re.sub("[ \t]{2,}", " ", text)
        paragraphs = [
            textwrap.fill(p, width, initial_indent=indent, subsequent_indent=indent)
            for p in re.split("\n\n", text)
        ]
        return "\n\n".join(paragraphs)


T = TypeVar("T")


class CompositeType:
    """A composite of multiple data types.
    """

    def __init__(self, *types):
        self.types = types

    def __call__(self, string):
        result = string
        for datatype in self.types:
            result = datatype(result)
        return result


class ComparisonValidator(Generic[T]):
    """
    Validator that compares an argument against an expected value.
    """

    def __init__(self, rhs: T, oper: Callable[[T, T], bool], expected: bool = True):
        self.rhs = rhs
        self.oper = oper
        self.expected = expected

    def __call__(self, lhs: T):
        if self.oper(lhs, self.rhs) != self.expected:
            raise ArgumentTypeError(
                f"{self.oper}({lhs}, {self.rhs}) != {self.expected}"
            )

        return lhs


class CharList:
    """
    Parses a string into a list of characters and ensures they are all in the choices
    tuple.
    """

    def __init__(self, choices):
        self.choices = set(choices)

    def __call__(self, string):
        chars = list(string)

        if not all(char in self.choices for char in chars):
            raise ArgumentTypeError(
                f"One or more characters in {chars} not in {self.choices}"
            )

        return chars


class Delimited(Generic[T]):
    """
    Splits a string argument using a delimiter.
    """

    def __init__(
        self,
        delim: str = ",",
        data_type: Optional[Callable[[str], T]] = None,
        choices: Optional[Sequence[Union[str, T]]] = None,
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
    ):
        self.delim = delim
        self.data_type = data_type
        self.choices = choices
        self.min_len = min_len
        self.max_len = max_len

    def __call__(self, arg: Union[str, List[T]]) -> List[T]:
        if isinstance(arg, str):
            vals = arg.split(self.delim) if self.delim else [arg]
        else:
            vals = arg
        if vals[0] == "*" and self.choices is not None:
            vals = self.choices
        if self.data_type:
            vals = [self.data_type(v) for v in vals]

        if self.min_len and len(vals) < self.min_len:
            raise ArgumentTypeError(f"There must be at least {self.min_len} values")

        if self.max_len and len(vals) > self.max_len:
            raise ArgumentTypeError(f"There can be at most {self.max_len} values")

        return vals


class EnumChoice(Generic[T]):
    def __init__(self, enum_class: Callable[[str], T] = None):
        self.enum_class = enum_class

    def __call__(self, arg: str) -> T:
        return self.enum_class(arg)


class AccessiblePath:
    """
    Tests that a path is accessible.
    """

    def __init__(self, path_type: PathType, mode: Permission):
        self.path_type = path_type
        self.mode = mode

    def __call__(self, path: Union[str, PathLike]) -> Path:
        path = as_path(path)

        if self.path_type == PathType.FILE and path in (STDOUT, STDERR):
            return path

        try:
            return Path(check_path(path, self.path_type, self.mode))
        except IOError:
            raise ArgumentTypeError(
                f"Path {path} is not an accessible {self.path_type} with mode "
                f"{self.mode}"
            )


def existing_path(path: Union[str, PathLike]) -> Path:
    """
    Tests that a path exists.
    """
    path = as_path(path)

    if path == STDOUT:
        return path

    try:
        return Path(resolve_path(path))
    except IOError:
        raise ArgumentTypeError(f"Path {path} does not exist")


def readable_file(path):
    try:
        return check_readable_file(resolve_path(path))
    except IOError:
        raise ArgumentTypeError(
            f"Path {path} does not exist, is not a file, or is not readable"
        )


def writable_file(path):
    try:
        return check_writable_file(as_path(path, "w"))
    except IOError:
        raise ArgumentTypeError(f"Path {path} is not writable")


def readwritable_file(path):
    """Tests that a file is both readable and writeable."""
    path = as_path(path)

    if path.exists():
        path = readable_file(path)

    return writable_file(path)


def readable_url(url):
    """
    Validator for a URL that must be readable.

    Args:
        url: The URL to validate.
    """
    parsed = urlparse(url)
    scheme = parsed.scheme or "file"
    if scheme == "file":
        filename = readable_file(parsed.path)
        return "file:" + str(filename)
    else:
        return url


str_list = Delimited[str](data_type=str)
"""Comma-delimited list of strings."""

INT_OR_STR_RE = re.compile(r"([\d.]+)([KkMmGg]?)")


def int_or_str(arg):
    """
    Similar to int(), but accepts K, M, and G abbreviations.
    """
    if arg is None or isinstance(arg, int):
        return arg
    elif isinstance(arg, str):
        match = INT_OR_STR_RE.match(arg.upper())
        num, mult = match.groups()
        if mult:
            return int(float(num) * Magnitude[mult].divisor)
        else:
            return int(num)
    else:
        raise ArgumentTypeError(f"Unsupported type {arg}")


def positive(type_=int, inclusive=False):
    """
    Tests that a number is greater than (or equal to, if ``inclusive=True``) zero.
    """
    oper = operator.ge if inclusive else operator.gt
    return CompositeType(type_, ComparisonValidator(0, oper))


N = TypeVar("N", bound=Number)


def between(min_val: N = None, max_val: N = None, type_: Type[N] = int):
    """
    Returns a CompositeType that validates `min_val <= x <= max_val`.
    """
    return CompositeType(
        type_,
        ComparisonValidator(min_val, operator.ge),
        ComparisonValidator(max_val, operator.le),
    )


probability = between(0, 1, float)
"""A float between 0-1 (inclusive)."""



from argparse import Namespace
from itertools import repeat
from pathlib import Path
from typing import IO, Optional, Sequence

from xphyle import open_

from atropos.commands.legacy_reports import Printer, TitlePrinter
from atropos.commands.reports import BaseReportGenerator, ReportWriter, BaseReportWriter
from atropos.io.formatters import FastaFormat


class DetectTextReportWriter(BaseReportWriter):
    def __init__(self, options: Namespace):
        super().__init__("txt", options)

    def serialize(self, summary: dict, stream: IO) -> None:
        names = summary["input"]["input_names"] or repeat(None)
        n_reads = summary["record_counts"][0]
        for input_idx, (matches, name) in enumerate(
            zip(summary["detect"]["matches"], names), 1
        ):
            self._generate_detector_report(stream, input_idx, n_reads, matches, name)

    @staticmethod
    def _generate_detector_report(
        stream: IO,
        input_idx: int,
        n_reads: int,
        matches: Sequence[dict],
        input_name: Optional[str] = None
    ):
        n_matches = len(matches)
        pad_size = len(str(n_matches))
        _print = Printer(stream)
        _print_title = TitlePrinter(stream)
        _print_indent = Printer(stream, indent=" " * (pad_size + 2))
        _print.newline()
        _print_title(f"Input {input_idx}", level=0)
        if input_name:
            _print(f"File: {input_name}")
        _print(f"Detected {n_matches} adapters/contaminants:")

        if n_matches == 0:
            _print("Try increasing --max-reads")
            return

        for idx, match in enumerate(matches):
            _print(
                ("{:>" + str(pad_size) + "}. Longest kmer: {}").format(
                    idx + 1, match["longest_kmer"]
                )
            )
            if match["longest_match"]:
                _print_indent(
                    f"Longest matching sequence: {match['longest_match']}"
                )
            if match["is_known"]:
                _print_indent(
                    "Name(s): {}".format(
                        f",\n{' ' * (pad_size + 11)}".join(match["known_names"])
                    )
                )
                _print_indent(
                    "Known sequence(s): {}".format(
                        f",\n{' ' * (pad_size + 11)}".join(match["known_seqs"])
                    )
                )
                _print_indent(
                    f"Known sequence K-mers that match detected contaminant: "
                    f"{match['known_to_contaminant_match_frac']:.2%}"
                )
            if match["abundance"]:
                _print_indent(
                    f"Abundance (full-length) in {n_reads} reads: {match['abundance']} "
                    f"({match['abundance'] / n_reads:.1%})"
                )
            if match["contaminant_to_known_match_frac"]:
                _print_indent(
                    f"Detected contaminant kmers that match known sequence: "
                    f"{match['contaminant_to_known_match_frac']:.2%}"
                )
            if match["kmer_freq_type"] == "frequency":
                _print_indent(f"Frequency of k-mers: {match['kmer_freq']:.2%}")
            else:
                _print_indent(f"Number of k-mer matches: {match['kmer_freq']}")


class DetectFastaReportWriter(ReportWriter):
    """

    """
    def __init__(self, options: Namespace):
        super().__init__("fasta", options)
        self.union = False
        if options.fasta:
            self.perinput = False
            for arg in options.fasta:
                setattr(self, arg, True)
        else:
            self.perinput = True

    def write_report(self, summary: dict, output_file: Path):
        names = summary["input"]["input_names"] or repeat(None)
        n_reads = summary["record_counts"][0]
        fasta_format = FastaFormat()
        union_records = []

        def format_match(_idx, _match, _records):
            name2 = [
                f"kmer_freq={_match['kmer_freq']}",
                f"kmer_freq_type={_match['kmer_freq_type']}",
            ]

            if _match["abundance"]:
                name2.append(f"abundance={_match['abundance']}")
                name2.append(f"abundance_frac={_match['abundance'] / n_reads}")

            if _match["contaminant_to_known_match_frac"]:
                name2.append(
                    f"contaminant_to_known_match_frac="
                    f"{_match['contaminant_to_known_match_frac']}"
                )

            if _match["is_known"]:
                name1 = _match["known_names"][0]
                name3 = []

                if len(_match["known_names"]) > 1:
                    name3 = [
                        f"other_names={'|'.join(_match['known_names'][1:])}"
                    ]

                if len(_match["known_seqs"]) > 1:
                    for seq in _match["known_seqs"]:
                        _records.append(
                            fasta_format.format_entry(
                                f"{name1}.{_idx} {';'.join(name2 + name3)}", seq
                            )
                        )
                else:
                    _records.append(
                        fasta_format.format_entry(
                            f"{name1} {';'.join(name2 + name3)}",
                            _match["known_seqs"][0],
                        )
                    )
            else:
                _records.append(
                    fasta_format.format_entry(
                        f"{_idx} {';'.join(name2)}", _match["longest_kmer"]
                    )
                )

        for i, (name, matches) in enumerate(zip(names, summary["detect"]["matches"])):
            records = []
            for idx, match in enumerate(matches, 1):
                format_match(idx, match, records)
            if self.union:
                union_records.extend(records)
            if self.perinput:
                path = self._get_output_file(output_file, (f"{i}.fasta",))
                with open_(path, "wt") as out:
                    out.write("".join(records))

        if self.union:
            with open_(self._get_output_file(output_file), "wt") as stream:
                stream.write("".join(union_records))


class DetectReportGenerator(BaseReportGenerator):
    @classmethod
    def _create_report_writer(cls, fmt: str, options: Namespace) -> ReportWriter:
        if fmt == "txt":
            return DetectTextReportWriter(options)
        elif fmt == "fasta":
            return DetectFastaReportWriter(options)
        else:
            return super()._create_report_writer(fmt, options)



from contextlib import contextmanager
from importlib import import_module
from io import BufferedWriter, BytesIO, TextIOWrapper
import os
from pathlib import Path
import sys
import traceback
from unittest.mock import patch
import urllib.request

from atropos.commands.trim.console import TrimCommandConsole


@contextmanager
def redirect_stderr():
    """Send stderr to stdout. Nose doesn't capture stderr, yet."""
    old_stderr = sys.stderr
    sys.stderr = sys.stdout
    yield

    sys.stderr = old_stderr


def datapath(path):
    return Path(__file__).parent / "data" / path


def datapathstr(path):
    return str(datapath(path))


def dataurl(path):
    return f"file:{datapath(path)}"


def cutpath(path):
    return Path(__file__).parent / "cut" / path


def assert_files_equal(path1, path2):
    # return os.system("diff -u {0} {1}".format(path1, path2)) == 0
    #with open(path1, "r") as i1, open(path2, "r") as i2:
    #    print("<[{}]>".format(i1.read()))
    #    print("<[{}]>".format(i2.read()))

    from subprocess import check_output, CalledProcessError

    try:
        check_output("diff -u {0} {1}".format(path1, path2), shell=True)
    except CalledProcessError as e:
        raise AssertionError(
            f"Files not equal: {path1} != {path2}\n"
            f"Diff: <{e.output.decode('utf-8')}>"
        )


def run(
    params,
    expected,
    inpath=None,
    inpath2=None,
    qualfile=None,
    interleaved_input=False,
    interleaved_output=False,
    sra_accn=None,
    stdout=False,
    output_dir=None,
):
    if type(params) is str:
        params = params.split()

    tmp_fastaq = output_dir / expected

    if sra_accn:
        params += ["--accession", sra_accn]
    elif interleaved_input:
        params += ["-l", inpath]
    elif inpath2:
        params += ["-pe1", datapath(inpath)]
        params += ["-pe2", datapath(inpath2)]
    else:
        params += ["-se", datapath(inpath)]
        if qualfile:
            params += ["-sq", datapath(qualfile)]

    if stdout:
        # Output is going to stdout, so we need to redirect it to the
        # temp file
        with intercept_stdout() as stdout:
            # print(params)
            retcode, summary = TrimCommandConsole.execute(
                tuple(str(p) for p in params)
            )
            with open(tmp_fastaq, "wt") as out:
                out.write(stdout.getvalue())
    else:
        if interleaved_output:
            params += ["-L", tmp_fastaq]
        else:
            params += ["-o", tmp_fastaq]  # TODO not parallelizable
        # print(params)
        retcode, summary = TrimCommandConsole.execute(
            tuple(str(p) for p in params)
        )

    assert summary is not None
    assert isinstance(summary, dict)

    if "exception" in summary and summary["exception"] is not None:
        assert retcode != 0
        err = summary["exception"]
        traceback.print_exception(*err["details"])
        raise Exception("Unexpected error: {}".format(err["message"]))
    else:
        assert retcode == 0

    # TODO redirect standard output
    assert os.path.exists(tmp_fastaq)
    assert_files_equal(cutpath(expected), tmp_fastaq)


# TODO diff log files
def approx_equal(a, b, tol):
    return abs(a - b) <= tol


def no_internet(url="https://github.com"):
    """Test whether there's no internet connection available.
    """
    try:
        urllib.request.urlopen(url).info()
        return False
    except:
        return True


def no_import(lib):
    """Test whether a library is importable.
    """
    try:
        mod = import_module(lib)
        return mod is None
    except ImportError:
        return True


class MockStdout:
    def __init__(self, name, as_bytes):
        self.bytes_io = BytesIO()
        object.__setattr__(self.bytes_io, "name", name)
        self.wrapper = TextIOWrapper(BufferedWriter(self.bytes_io))
        self.as_bytes = as_bytes

    def getvalue(self):
        self.wrapper.flush()
        val = self.bytes_io.getvalue()
        if not self.as_bytes:
            val = val.decode()
        return val


@contextmanager
def intercept_stdout(as_bytes=False):
    i = MockStdout("<stdout>", as_bytes)
    with patch("sys.stdout", i.wrapper):
        yield i


@contextmanager
def intercept_stderr(as_bytes=False):
    i = MockStdout("<stderr>", as_bytes)
    with patch("sys.stderr", i.wrapper):
        yield i



import sys
from typing import Optional, Sequence

from atropos.console import run_atropos


def main(args: Optional[Sequence[str]] = None):
    """
    Main method, designed to be called by the system.

    Always calls `sys.exit()`. If you want to run Atropos programatically, call
    `run()` instead.

    Args:
        args: Command-line arguments. If None, `sys.argv[1:]` is used.
    """
    sys.exit(run_atropos(args))


if __name__ == "__main__":
    main()



# kate: syntax Python;
# cython: profile=False, emit_code_comments=False
from atropos.errors import FormatError
from atropos.utils import classproperty

from ._base import PrefetchSequenceReader, estimate_num_records


class FastqReader(PrefetchSequenceReader):
    """
    Reader for FASTQ files. Does not support multi-line FASTQ files.
    """
    @classproperty
    def file_format(cls) -> str:
        return "FASTQ"

    @classproperty
    def delivers_qualities(cls) -> bool:
        return True

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def colorspace(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return False

    def estimate_num_records(self):
        return estimate_num_records(self.name, self._first_seq.size_in_bytes, 4, 2)

    def _iter(self):
        """
        Yield Sequence objects
        """
        cdef int i = 0
        cdef int strip
        cdef str line, name, qualities, sequence, name2
        sequence_factory = self._sequence_factory

        it = iter(self._file)
        line = next(it)
        if not (line and line[0] == '@'):
            raise FormatError(
                f"Line {i+1} in FASTQ file is expected to start with '@', but found "
                f"{line[:10]!r}"
            )
        strip = -2 if line.endswith('\r\n') else -1
        name = line[1:strip]

        i = 1
        for line in it:
            if i == 0:
                if not (line and line[0] == '@'):
                    raise FormatError(
                        "Line {i+1} in FASTQ file is expected to start with '@', but "
                        "found {line[:10]!r}"
                    )
                name = line[1:strip]
            elif i == 1:
                sequence = line[:strip]
            elif i == 2:
                if line == '+\n':  # check most common case first
                    name2 = ''
                else:
                    line = line[:strip]
                    if not (line and line[0] == '+'):
                        raise FormatError(
                            f"Line {i+1} in FASTQ file is expected to start with '+', "
                            f"but found {line[:10]!r}"
                        )
                    if len(line) > 1:
                        if not line[1:] == name:
                            raise FormatError(
                                f"At line {i+1}: Sequence descriptions in the FASTQ "
                                f"file don't match ({name!r} != {line[1:]!r}).\n"
                                f"The second sequence description must be either empty "
                                f"or equal to the first description."
                            )
                        name2 = name
                    else:
                        name2 = ''
            elif i == 3:
                if len(line) == len(sequence) - strip:
                    qualities = line[:strip]
                else:
                    qualities = line.rstrip('\r\n')
                try:
                    yield sequence_factory(
                        name, sequence, qualities, name2=name2, alphabet=self._alphabet
                    )
                except Exception as err:
                    raise FormatError(
                        f"Error creating sequence record at line {i+1}"
                    ) from err
            i = (i + 1) % 4
        if i != 0:
            raise FormatError("FASTQ file ended prematurely")



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import List, Optional, Tuple, Union

from atropos.errors import UnknownFileTypeError
from atropos.io import guess_format_from_name
from atropos.io.sequence import Sequence, ColorspaceSequence


class SequenceFileFormat(metaclass=ABCMeta):
    """
    Base class for sequence formatters.
    """

    @abstractmethod
    def format(self, read: Sequence) -> str:
        """Formats a Sequence as a string.

        Args:
            read: The Sequence object.

        Returns:
            A string representation of the sequence object in the sequence
            file format.
        """

    def header(self) -> Optional[str]:
        """
        Returns the header to be written at the beginning of the file.
        """


class FastaFormat(SequenceFileFormat):
    """
    FASTA SequenceFileFormat.
    """

    def __init__(self, line_length=None):
        """
        Args:
            line_length: Max line length (in characters), or None. Determines
                whether and how lines are wrapped.
        """
        if line_length:
            from textwrap import TextWrapper

            self._text_wrapper = TextWrapper(width=line_length)
        else:
            self._text_wrapper = None

    def format(self, read: Sequence) -> str:
        return self.format_entry(read.name, read.sequence)

    def format_entry(self, name: str, sequence: str) -> str:
        """
        Converts a sequence record to a string.
        """
        if self._text_wrapper:
            sequence = self._text_wrapper.fill(sequence)

        return "".join((">", name, "\n", sequence, "\n"))


class ColorspaceFastaFormat(FastaFormat):
    """
    FastaFormat in which sequences are in colorspace.
    """

    def format(self, read: ColorspaceSequence) -> str:
        return self.format_entry(read.name, read.primer + read.sequence)


class FastqFormat(SequenceFileFormat):
    """
    FASTQ SequenceFileFormat.
    """

    def format(self, read: Sequence) -> str:
        return self.format_entry(read.name, read.sequence, read.qualities, read.name2)

    @classmethod
    def format_entry(
        cls, name: str, sequence: str, qualities: str, name2: Optional[str] = None
    ) -> str:
        """
        Converts a sequence record to a string.
        """
        return "".join(
            ("@", name, "\n", sequence, "\n+", name2 or "", "\n", qualities, "\n")
        )


class ColorspaceFastqFormat(FastqFormat):
    """
    FastqFormat in which sequences are in colorspace.
    """

    def format(self, read: ColorspaceSequence) -> str:
        return self.format_entry(
            read.name, read.primer + read.sequence, read.qualities, read.name2
        )


class SAMFormat(SequenceFileFormat):
    """
    SAM SequenceFileFormat.

    TODO: when the input is SAM/BAM and the output is SAM, propagate the header to
     the output file.
    """

    def __init__(self, flag: int):
        self._flag = str(flag)

    def format(self, read: Sequence) -> str:
        return "".join(
            (
                read.name,
                "\t",
                self._flag,
                "\t*\t0\t0\t*\t*\t0\t0\t",
                read.sequence,
                "\t",
                read.qualities,
                "\n",
            )
        )

    def header(self) -> str:
        return "@HD\tVN:1.6\tSO:unsorted\n"


class Formatter(metaclass=ABCMeta):
    """
    Base class for Formatters.
    """

    def __init__(self):
        self.written = 0
        self.read1_bp = 0
        self.read2_bp = 0

    @abstractmethod
    def format(
        self, result: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        """
        Formats read(s) and add them to `result`.

        Args:
            result: A dict mapping file names to lists of formatted reads.
            read1: The first read to format.
            read2: The second read to format.
        """

    @classmethod
    def _get_result_list(
        cls,
        result: dict,
        file1: str,
        seq_format1: SequenceFileFormat,
        seq_format2: Optional[SequenceFileFormat] = None
    ) -> List[str]:
        if file1 in result:
            return result[file1]
        else:
            result_list = []
            header = seq_format1.header()

            if header:
                if seq_format2 and header != seq_format2.header():
                    raise ValueError("sequence formats must have same header")

                result_list.append(header)

            result[file1] = result_list

            return result_list

    @property
    def written_bp(self) -> Tuple[int, int]:
        """
        Tuple of base-pairs written (read1_bp, read2_bp).
        """
        return self.read1_bp, self.read2_bp


class SingleEndFormatter(Formatter):
    """
    Wrapper for a SequenceFileFormat for single-end data.
    """

    def __init__(self, file1: str, seq_format: SequenceFileFormat):
        """
        Args:
            file1: The single-end file.
            seq_format: The SequenceFileFormat object.
        """
        super().__init__()
        self._file1 = file1
        self._seq_format = seq_format

    def format(
        self, result: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        """
        Formats read(s) and add them to `result`.

        Args:
            result: A dict mapping file names to lists of formatted reads.
            read1:
            read2: The reads to format.
        """
        result_list = self._get_result_list(result, self._file1, self._seq_format)
        result_list.append(self._seq_format.format(read1))
        self.written += 1
        self.read1_bp += len(read1)


class InterleavedFormatter(Formatter):
    """
    Formats read pairs as successive reads in an interleaved file.
    """

    def __init__(
        self,
        file1: str,
        seq_format1: SequenceFileFormat,
        seq_format2: Optional[SequenceFileFormat] = None
    ):
        super().__init__()
        self._file1 = file1
        self._seq_format1 = seq_format1
        self._seq_format2 = seq_format2 or seq_format1

    def format(
        self, result: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        result_list = self._get_result_list(
            result, self._file1, self._seq_format1, self._seq_format2
        )
        result_list.append(self._seq_format1.format(read1))
        result_list.append(self._seq_format2.format(read2))
        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)


class PairedEndFormatter(Formatter):
    """
    Wrapper for a SequenceFileFormat. Both reads in a pair are formatted using the
    specified format.
    """

    def __init__(
        self,
        file1: str,
        file2: str,
        seq_format1: SequenceFileFormat,
        seq_format2: Optional[SequenceFileFormat] = None
    ):
        super().__init__()
        self._data = [(file1, seq_format1), (file2, seq_format2 or seq_format1)]

    def format(
        self, result: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        for read, (path, seq_format) in zip((read1, read2), self._data):
            result_list = self._get_result_list(result, path, seq_format)
            result_list.append(seq_format.format(read))

        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)


class SingleEndSAMFormatter(SingleEndFormatter):
    def __init__(self, file1: str):
        super().__init__(file1, SAMFormat(0))


class PairedEndSAMFormatter(InterleavedFormatter):
    def __init__(self, file1: str):
        super().__init__(file1, SAMFormat(65), SAMFormat(129))


def sra_colorspace_sequence(
    name: str, sequence: str, qualities: str, name2: str
):
    """
    Factory for an SRA colorspace sequence (which has one quality value too many).
    """
    return ColorspaceSequence(name, sequence, qualities[1:], name2=name2)


def create_seq_formatter(
    file1: Union[str, Path],
    file2: Optional[Union[str, Path]] = None,
    qualities: Optional[bool] = None,
    colorspace: bool = False,
    file_format: Optional[Union[str, SequenceFileFormat]] = None,
    interleaved: bool = False,
    line_length: Optional[int] = None,
):
    """
    Creates a Formatter, deriving the format name from the file extension.

    Args:
        file1: First output file.
        file2: Second output file.
        qualities: When file_format is None, this can be set to True or False to
            specify whether the written sequences will have quality values. This is
            is used in two ways:
            * If the output format cannot be determined (unrecognized extension
              etc), no exception is raised, but fasta or fastq format is chosen
              appropriately.
            * When False (no qualities available), an exception is raised when
              the auto-detected output format is FASTQ.
        colorspace: If True, instances of the Colorspace... formats are returned.
        file_format: If set to None, file format is autodetected from the file
            name extension. Set to 'fasta', 'fastq', or 'sra-fastq' to not
            auto-detect. Colorspace is not auto-detected and must always be
            requested explicitly.
        interleaved: Whether the output should be interleaved (file2 must be None).
        line_length: Maximum length of a sequence line in FASTA output.

    Returns:
        A Formatter instance.
    """
    if file_format is None:
        file_format = guess_format_from_name(file1, raise_on_failure=qualities is None)

    if file_format is not None:
        file_format = file_format.lower()
    elif qualities is True:
        # Format not recognized, but know we want to write reads with qualities.
        file_format = "fastq"
    elif qualities is False:
        # Same, but we know that we want to write reads without qualities.
        file_format = "fasta"
    else:
        raise UnknownFileTypeError("Could not determine file type.")

    if file_format in ("sam", "fastq") and qualities is False:
        raise ValueError(
            f"Output format cannot be {file_format} since no quality values are "
            f"available."
        )

    if file_format == "sam":
        if file2 is not None:
            raise ValueError("Only one output file allowed for SAM format")

        if interleaved:
            return PairedEndSAMFormatter(file1)
        else:
            return SingleEndSAMFormatter(file1)
    else:
        if file_format == "fasta":
            if colorspace:
                fmt = ColorspaceFastaFormat(line_length)
            else:
                fmt = FastaFormat(line_length)
        elif file_format == "fastq":
            if colorspace:
                fmt = ColorspaceFastqFormat()
            else:
                fmt = FastqFormat()
        else:
            raise UnknownFileTypeError(
                f"File format {file_format!r} is unknown (expected 'fasta' or "
                f"'fastq')."
            )

        if file2 is not None:
            return PairedEndFormatter(file1, file2, fmt)
        elif interleaved:
            return InterleavedFormatter(file1, fmt)
        else:
            return SingleEndFormatter(file1, fmt)



import os
import shutil
from typing import Callable, Iterable, Union

from pytest import raises
from xphyle import open_

from atropos.console import execute_cli
from atropos.commands.trim.console import TrimCommandConsole
from atropos.utils import ReturnCode

from .utils import (
    run,
    assert_files_equal,
    datapath,
    cutpath,
    redirect_stderr,
    intercept_stdout,
)

BACK_ALIGNERS = ("adapter", "insert")


def run_paired(
    params: Union[str, list],
    in1: str,
    in2: str,
    expected1: str,
    expected2: str,
    aligners: Iterable[str] = ("adapter",),
    callback: Callable = None,
    files_equal: bool = True,
    error_on_rc: bool = True,
    output_dir=None,
):
    if type(params) is str:
        params = params.split()

    for aligner in aligners:
        p1 = output_dir / f"tmp1-{expected1.format(aligner=aligner)}"
        p2 = output_dir / f"tmp2-{expected2.format(aligner=aligner)}"
        p = params.copy()
        p += ["--aligner", aligner, "-o", str(p1), "-p", str(p2)]

        infiles = []

        for i, infile in enumerate((in1, in2), 1):
            inpath = datapath(infile.format(aligner=aligner))
            infiles.append(inpath)
            p.append(f"-pe{i}")
            p.append(str(inpath))

        result = TrimCommandConsole.execute(p)

        assert isinstance(result, tuple)
        assert len(result) == 2

        if error_on_rc:
            err = (
                result[1]["exception"]
                if result[1] and "exception" in result[1]
                else None
            )
            if result[0] != 0:
                if err is None:
                    raise AssertionError("Return code {} != 0".format(result[0]))
                else:
                    raise AssertionError(
                        "Return code {} != 0".format(result[0])
                    ) from err["details"][1]

        if files_equal:
            assert_files_equal(cutpath(expected1.format(aligner=aligner)), p1)
            assert_files_equal(cutpath(expected2.format(aligner=aligner)), p2)

        if callback:
            callback(aligner, infiles, (p1, p2), result)


def run_interleaved(
    params: Union[str, list],
    inpath: str,
    expected: str,
    aligners: Iterable[str] = ("adapter",),
    stdout: bool = False,
    error_on_rc: bool = True,
    output_dir=None,
):
    if type(params) is str:
        params = params.split()

    for aligner in aligners:
        tmp = output_dir / expected.format(aligner=aligner)
        p = params.copy()
        p += ["--aligner", aligner, "-l", str(datapath(inpath.format(aligner=aligner)))]

        if stdout:
            # Output is going to stdout, so we need to redirect it to the
            # temp file
            with intercept_stdout() as stdout:
                # print(params)
                result = TrimCommandConsole.execute(p)
                with open(tmp, "wt") as out:
                    out.write(stdout.getvalue())
        else:
            p.extend(["-L", str(tmp)])
            result = TrimCommandConsole.execute(p)

        assert isinstance(result, tuple)
        assert len(result) == 2

        if error_on_rc:
            err = (
                result[1]["exception"]
                if result[1] and "exception" in result[1]
                else None
            )
            if result[0] != 0:
                raise AssertionError("Return code {} != 0".format(result[0])) from err[
                    "details"
                ][1]

        assert_files_equal(cutpath(expected.format(aligner=aligner)), tmp)


# def run_interleaved2(params, inpath, expected1, expected2, aligners=('adapter',)):
#     assert False  # unused function
#     if type(params) is str:
#         params = params.split()
#     for aligner in aligners:
#         with temporary_path('tmp1-' + expected1.format(aligner=aligner)) as p1:
#             with temporary_path('tmp2-' + expected2.format(aligner=aligner)) as p2:
#                 p = params.copy()
#                 p += ['--aligner', aligner, '--interleaved', '-o', p1, '-p', p2]
#                 p += [datapath(inpath.format(aligner=aligner))]
#                 assert execute_cli(p) == 0
#                 assert files_equal(cutpath(expected.format(aligner=aligner)), p1)
#                 assert files_equal(cutpath(expected.format(aligner=aligner)), p2)


def test_paired_separate(tmp_path):
    """test separate trimming of paired-end reads"""
    run(
        "-a TTAGACATAT",
        "paired-separate.1.fastq",
        "paired.1.fastq",
        output_dir=tmp_path,
    )
    run(
        "-a CAGTGGAGTA",
        "paired-separate.2.fastq",
        "paired.2.fastq",
        output_dir=tmp_path,
    )


def test_paired_end_legacy(tmp_path):
    """--paired-output, not using -A/-B/-G"""
    # The -m 14 filters out one read, which should then also be filtered out in
    # the second output file.
    run_paired(
        "-a TTAGACATAT -m 14",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired.m14.1.fastq",
        expected2="paired.m14.2.fastq",
        output_dir=tmp_path,
    )


def test_untrimmed_paired_output(tmp_path):
    untrimmed1 = tmp_path / "tmp-untrimmed.1.fastq"
    untrimmed2 = tmp_path / "tmp-untrimmed.2.fastq"

    def callback(aligner, infiles, outfiles, result):
        assert_files_equal(cutpath("paired-untrimmed.1.fastq"), untrimmed1)
        assert_files_equal(cutpath("paired-untrimmed.2.fastq"), untrimmed2)

    run_paired(
        [
            "-a",
            "TTAGACATAT",
            "--untrimmed-output",
            str(untrimmed1),
            "--untrimmed-paired-output",
            str(untrimmed2),
        ],
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired-trimmed.1.fastq",
        expected2="paired-trimmed.2.fastq",
        callback=callback,
        output_dir=tmp_path,
    )


def test_explicit_format_with_paired(tmp_path):
    # Use --format=fastq with input files whose extension is .txt
    txt1 = tmp_path / "paired.1.txt"
    txt2 = tmp_path / "paired.2.txt"
    shutil.copyfile(datapath("paired.1.fastq"), txt1)
    shutil.copyfile(datapath("paired.2.fastq"), txt2)
    run_paired(
        "--input-format=fastq -a TTAGACATAT -m 14",
        in1=str(txt1),
        in2=str(txt2),
        expected1="paired.m14.1.fastq",
        expected2="paired.m14.2.fastq",
        output_dir=tmp_path,
    )


def test_no_trimming_legacy():
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            "-a",
            "XXXXX",
            "-o",
            "/dev/null",
            "-p",
            "/dev/null",
            "-pe1",
            datapath("paired.1.fastq"),
            "-pe2",
            datapath("paired.2.fastq"),
        ]
    )


def test_no_trimming():
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            "-a",
            "XXXXX",
            "-A",
            "XXXXX",
            "-o",
            "/dev/null",
            "-p",
            "/dev/null",
            "-pe1",
            datapath("paired.1.fastq"),
            "-pe2",
            datapath("paired.2.fastq"),
        ]
    )


def test_missing_file():
    with redirect_stderr():
        assert ReturnCode.ERROR == execute_cli(
            ["-a", "XX", "--paired-output", "out.fastq", datapath("paired.1.fastq")]
        )


def test_first_too_short(tmp_path):
    trunc1 = tmp_path / "truncated.1.fastq"
    out = tmp_path / "out.fastq"
    # Create a truncated file in which the last read is missing
    with open_(datapath("paired.1.fastq")) as f:
        lines = f.readlines()
        lines = lines[:-4]
    with open_(trunc1, "w") as f:
        f.writelines(lines)
    with redirect_stderr():
        assert ReturnCode.ERROR == execute_cli([
            "-a", "XX", "--paired-output", str(out),
            str(trunc1), str(datapath("paired.2.fastq"))
        ])


def test_second_too_short(tmp_path):
    trunc2 = tmp_path / "truncated.2.fastq"
    out = tmp_path / "out.fastq"
    # Create a truncated file in which the last read is missing
    with open_(datapath("paired.2.fastq")) as f:
        lines = f.readlines()
        lines = lines[:-4]
    with open_(trunc2, "w") as f:
        f.writelines(lines)
    with redirect_stderr():
        assert ReturnCode.ERROR == execute_cli([
            "-a", "XX", "--paired-output", str(out),
            str(datapath("paired.1.fastq")), str(trunc2)
        ])


def test_unmatched_read_names(tmp_path):
    swapped = tmp_path / "swapped.1.fastq"
    out1 = tmp_path / "out1.fastq"
    out2 = tmp_path / "out2.fastq"

    # Create a file in which reads 2 and are swapped
    with open_(datapath("paired.1.fastq")) as f:
        lines = f.readlines()
        lines = lines[0:4] + lines[8:12] + lines[4:8] + lines[12:]
    with open_(swapped, "w") as f:
        f.writelines(lines)
    with redirect_stderr():
        result = TrimCommandConsole.execute([
            "-a", "XX", "-o", str(out1), "-p", str(out2),
            "-pe1", str(swapped), "-pe2", str(datapath("paired.2.fastq"))
        ])
        assert isinstance(result, tuple)
        assert len(result) == 2
        assert result[0] != 0


def test_legacy_minlength(tmp_path):
    """Ensure -m is not applied to second read in a pair in legacy mode"""
    run_paired(
        "-a XXX -m 27",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired-m27.1.fastq",
        expected2="paired-m27.2.fastq",
        output_dir=tmp_path,
    )


def test_paired_end(tmp_path):
    """single-pass paired-end with -m"""
    run_paired(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired_{aligner}.1.fastq",
        expected2="paired_{aligner}.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_paired_anchored_back_no_indels(tmp_path):
    run_paired(
        "-a BACKADAPTER$ -A BACKADAPTER$ -N --no-indels",
        in1="anchored-back.fasta",
        in2="anchored-back.fasta",
        expected1="anchored-back.fasta",
        expected2="anchored-back.fasta",
        output_dir=tmp_path,
    )


def test_paired_end_qualtrim(tmp_path):
    """single-pass paired-end with -q and -m"""
    run_paired(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="pairedq.1.fastq",
        expected2="pairedq.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_paired_end_qualtrim_swapped(tmp_path):
    """single-pass paired-end with -q and -m, but files swapped"""
    run_paired(
        "-q 20 -a CAGTGGAGTA -A TTAGACATAT -m 14 --adapter-max-rmp 0.001",
        in1="paired.2.fastq",
        in2="paired.1.fastq",
        expected1="pairedq.2.fastq",
        expected2="pairedq.1.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_paired_end_cut(tmp_path):
    run_paired(
        "-u 3 -u -1 -U 4 -U -2",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="pairedu.1.fastq",
        expected2="pairedu.2.fastq",
        output_dir=tmp_path,
    )


def test_paired_end_a_only(tmp_path):
    run_paired(
        "-A CAGTGGAGTA",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired-onlyA.1.fastq",
        expected2="paired-onlyA.2.fastq",
        output_dir=tmp_path,
    )


def test_paired_end_mask_adapter(tmp_path):
    """mask adapter with N (reads maintain the same length)"""
    run_paired(
        "-a CAAG -A TCGA -n 3 --mask-adapter",
        in1="back_repeat.1.fastq",
        in2="back_repeat.2.fastq",
        expected1="back_repeat.1.fastq",
        expected2="back_repeat.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_discard_untrimmed(tmp_path):
    # issue #146
    # the first adapter is a sequence cut out from the first read
    run_paired(
        "-a CTCCAGCTTAGACATATC -A XXXXXXXX --discard-untrimmed",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="empty.fastq",
        expected2="empty.fastq",
        output_dir=tmp_path,
    )


def test_discard_trimmed(tmp_path):
    run_paired(
        "-A C -O 1 --discard-trimmed",  # applies everywhere
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="empty.fastq",
        expected2="empty.fastq",
        output_dir=tmp_path,
    )


def test_interleaved(tmp_path):
    """single-pass interleaved paired-end with -q and -m"""
    run_interleaved(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        inpath="interleaved.fastq",
        expected="interleaved.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_interleaved_stdout(tmp_path):
    """single-pass interleaved paired-end with -q and -m"""
    run_interleaved(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        inpath="interleaved.fastq",
        expected="interleaved.fastq",
        aligners=BACK_ALIGNERS,
        stdout=True,
        output_dir=tmp_path
    )


def test_interleaved_no_paired_output(tmp_path):
    p1 = tmp_path / "temp-paired.1.fastq"
    p2 = tmp_path / "temp-paired.2.fastq"
    params = "-a XX --interleaved".split()
    with redirect_stderr():
        params += ["-o", p1, "-p1", p2, "paired.1.fastq", "paired.2.fastq"]
        assert execute_cli(params) == ReturnCode.ERROR


# TODO
# def test_interleaved_input_paired_output():
#     '''single-pass interleaved paired-end with -q and -m, paired output'''
#     run_interleaved2('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
#         inpath='interleaved.fastq', expected1='pairedq1.fastq',
#         expected2='pairedq2.fastq'
#     )
def test_pair_filter(tmp_path):
    run_paired(
        "--pair-filter=both -a TTAGACATAT -A GGAGTA -m 14",
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired-filterboth_{aligner}.1.fastq",
        expected2="paired-filterboth_{aligner}.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_too_short_paired_output(tmp_path):
    p1 = tmp_path / "temp-too-short.1.fastq"
    p2 = tmp_path / "temp-too-short.2.fastq"

    def callback(aligner, infiles, outfiles, result):
        assert_files_equal(cutpath("paired-too-short.1.fastq"), p1)
        assert_files_equal(cutpath("paired-too-short.2.fastq"), p2)

    run_paired(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output "
        "{0} --too-short-paired-output {1}".format(p1, p2),
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired_{aligner}.1.fastq",
        expected2="paired_{aligner}.2.fastq",
        aligners=BACK_ALIGNERS,
        callback=callback,
        output_dir=tmp_path,
    )


def test_too_long_output(tmp_path):
    p1 = tmp_path / "temp-too-long.1.fastq"
    p2 = tmp_path /"temp-too-long.2.fastq"

    def callback(aligner, infiles, outfiles, result):
        assert_files_equal(
            cutpath("paired_{aligner}.1.fastq".format(aligner=aligner)), p1
        )
        assert_files_equal(
            cutpath("paired_{aligner}.2.fastq".format(aligner=aligner)), p2
        )

    run_paired(
        "-a TTAGACATAT -A CAGTGGAGTA -M 14 --too-long-output "
        "{0} --too-long-paired-output {1}".format(p1, p2),
        in1="paired.1.fastq",
        in2="paired.2.fastq",
        expected1="paired-too-short.1.fastq",
        expected2="paired-too-short.2.fastq",
        aligners=BACK_ALIGNERS,
        callback=callback,
        output_dir=tmp_path,
    )


def test_too_short_output_paired_option_missing(tmp_path):
    with raises(SystemExit):
        p1 = tmp_path / "temp-too-short.1.fastq"
        run_paired(
            "-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output " "{0}".format(p1),
            in1="paired.1.fastq",
            in2="paired.2.fastq",
            expected1="paired.1.fastq",
            expected2="paired.2.fastq",
            aligners=BACK_ALIGNERS,
            output_dir=tmp_path,
        )


def test_custom_bisulfite_1(tmp_path):
    run_paired(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 -q 0 --bisulfite 2,2,1,1",
        in1="paired_bis_{aligner}.1.fastq",
        in2="paired_bis_{aligner}.2.fastq",
        expected1="paired_bis1_{aligner}.1.fastq",
        expected2="paired_bis1_{aligner}.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_custom_bisulfite_2(tmp_path):
    run_paired(
        "-a TTAGACATAT -A CAGTGGAGTA -m 10 -q 0 --bisulfite 20,20,1,1;0,0,0,0",
        in1="paired_bis_{aligner}.1.fastq",
        in2="paired_bis_{aligner}.2.fastq",
        expected1="paired_bis2_{aligner}.1.fastq",
        expected2="paired_bis2_{aligner}.2.fastq",
        aligners=BACK_ALIGNERS,
        output_dir=tmp_path,
    )


def test_no_insert_match(tmp_path):
    # with -O
    # Note: this fails if you set -e 0.3 because the higher error rate enables a
    # match of 7 bp, even though it has 2 errors. This illustrates why using
    # --adapter-max-rmp is better.
    # run_paired('-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -O 7 -m 25 -q 0 --trim-n',
    #     in1='insert.1.fastq', in2='insert.2.fastq',
    #     expected1='insert.1.fastq', expected2='insert.2.fastq',
    #     aligners=('insert',)
    # )
    # with --adapter-max-rmp
    run_paired(
        "-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -e 0.3 --adapter-max-rmp 0.001 -m 25 -q 0 --trim-n",
        in1="insert.1.fastq",
        in2="insert.2.fastq",
        expected1="insert.1.fastq",
        expected2="insert.2.fastq",
        aligners=("insert",),
        output_dir=tmp_path,
    )


def test_overwrite(tmp_path):
    run_paired(
        "-w 10,30,10",
        in1="lowq.fastq",
        in2="highq.fastq",
        expected1="lowq.fastq",
        expected2="highq.fastq",
        output_dir=tmp_path,
    )


def test_no_writer_process(tmp_path):
    def check_multifile(aligner, infiles, outfiles, result):
        assert os.path.basename(outfiles[0]) == "tmp1-out.1.fastq"
        assert os.path.basename(outfiles[1]) == "tmp2-out.2.fastq"
        tmpdir = os.path.dirname(outfiles[0])
        assert tmpdir == os.path.dirname(outfiles[1])
        # TODO: If the final worker doesn't get the chance to process any
        # batches, the last output file is never created.
        assert os.path.exists(os.path.join(tmpdir, "tmp1-out.1.0.fastq"))
        assert os.path.exists(os.path.join(tmpdir, "tmp1-out.1.1.fastq"))
        # assert os.path.exists(os.path.join(tmpdir, 'tmp1-out.1.2.fastq'))
        assert os.path.exists(os.path.join(tmpdir, "tmp2-out.2.0.fastq"))
        assert os.path.exists(os.path.join(tmpdir, "tmp2-out.2.1.fastq"))

    # assert os.path.exists(os.path.join(tmpdir, 'tmp2-out.2.2.fastq'))
    # TODO: check contents
    run_paired(
        "--threads 3 --no-writer-process --batch-size 1 -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        in1="big.1.fq",
        in2="big.2.fq",
        expected1="out.1.fastq",
        expected2="out.2.fastq",
        aligners=BACK_ALIGNERS,
        files_equal=False,
        callback=check_multifile,
        output_dir=tmp_path,
    )


def test_summary(tmp_path):
    def check_summary(aligner, infiles, outfiles, result):
        summary = result[1]
        assert summary is not None
        assert isinstance(summary, dict)
        assert summary["command"] == "trim"
        assert summary["options"]["orig_args"] == [
            "--threads",
            "2",
            "-a",
            "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG",
            "-A",
            "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
            "--aligner",
            aligner,
            "-o",
            str(outfiles[0]),
            "-p",
            str(outfiles[1]),
            "-pe1",
            str(infiles[0]),
            "-pe2",
            str(infiles[1]),
        ]
        assert summary["sample_id"] == "big"
        assert summary["mode"] == "parallel"
        assert summary["threads"] == 2
        # infiles are 100 125 bp PE reads
        expected_record_counts = {0: 100}
        assert summary["record_counts"] == expected_record_counts
        expected_bp_counts = {0: [12500, 12500]}
        assert summary["bp_counts"] == expected_bp_counts
        assert "timing" in summary
        assert "wallclock" in summary["timing"]
        assert summary["timing"]["wallclock"] > 0
        assert "cpu" in summary["timing"]
        assert summary["timing"]["cpu"] > 0

    run_paired(
        "--threads 2 -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        in1="big.1.fq",
        in2="big.2.fq",
        expected1="out.1.fastq",
        expected2="out.2.fastq",
        aligners=BACK_ALIGNERS,
        files_equal=False,
        callback=check_summary,
        output_dir=tmp_path,
    )


def test_sam(tmp_path):
    run_interleaved(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 --output-format sam",
        "paired.sam",
        "paired_insert.sam",
        aligners=["insert"],
        output_dir=tmp_path,
    )


# def test_long_reads():
#    run_paired(''
#        in1='long.1.fq', in2='long.2.fq',
#        expected1=, expected2=,
#        aligners=BACK_ALIGNERS)


def test_issue68(tmp_path):
    run_paired(
        "--error-rate 0.20 --insert-match-error-rate 0.30 --minimum-length 20 --aligner insert -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        in1="issue68.1.fq",
        in2="issue68.2.fq",
        expected1="issue68.1.fq",
        expected2="issue68.2.fq",
        aligners=["insert"],
        output_dir=tmp_path,
    )



import os
import tempfile

from atropos.adapters import Adapter, AdapterType, ColorspaceAdapter
from atropos.commands.trim.modifiers import AdapterCutter
from atropos.commands.trim.multicore import OrderPreservingWriterResultHandler
from atropos.commands.trim.writers import Writers
from atropos.io.sequence import ColorspaceSequence, Sequence


def test_cs_5p():
    read = ColorspaceSequence("name", "0123", "DEFG", "T")
    adapter = ColorspaceAdapter("CG", AdapterType.PREFIX, 0.1)
    cutter = AdapterCutter([adapter])
    cutter(read)


# no assertion here, just make sure the above code runs without
# an exception
def test_metrics():
    read = Sequence("name", "AAAACCCCAAAA")
    adapters = [Adapter("CCCC", AdapterType.BACK, 0.1)]
    cutter = AdapterCutter(adapters, times=3)
    cutter(read)
    # TODO make this a lot simpler
    trimmed_bp = 0
    for adapter in adapters:
        for d in (adapter.lengths_front, adapter.lengths_back):
            trimmed_bp += sum(seqlen * count for (seqlen, count) in d.items())
    assert trimmed_bp <= len(read), trimmed_bp


def test_order_preserving_writer():
    path = tempfile.mkstemp()[1]
    try:
        writers = Writers()
        handler = OrderPreservingWriterResultHandler(writers)
        handler.start(None)
        # write three batches out of order
        result2 = "result2"
        handler.write_result(2, {path: result2})
        result3 = "result3"
        handler.write_result(3, {path: result3})
        result1 = "result1"
        handler.write_result(1, {path: result1})
        handler.finish(total_batches=3)
        # check that the results are in the right order
        with open(path, "rt") as inp:
            assert inp.read() == (result1 + result2 + result3)
    finally:
        os.remove(path)



from multiprocessing import Process, Queue
import os
from typing import Dict, Iterable, Optional

from loguru import logger
import xphyle
from xphyle.formats import CompressionFormat

from atropos.commands.multicore import ParallelPipelineRunner, WorkerProcess
from atropos.commands.trim import (
    ResultHandler,
    WorkerResultHandler,
    WriterResultHandler,
    TrimCommand,
    TrimPipeline,
)
from atropos.commands.trim.writers import Writers
from atropos.utils import ReturnCode
from atropos.utils.multicore import (
    Control,
    ControlSignal,
    MulticoreError,
    PendingQueue,
    enqueue,
    dequeue,
    kill,
    wait_on_process,
)


class Done(MulticoreError):
    """
    Raised when process exits normally.
    """

    pass


class Killed(MulticoreError):
    """
    Raised when process exits is killed.
    """

    pass


class WriterManager:
    """
    Manager for a writer process and control variable.
    """

    def __init__(
        self,
        writers: Writers,
        compression_mode: str,
        preserve_order: bool,
        result_queue: Queue,
        timeout: int,
    ):
        # result handler
        if preserve_order:
            writer_result_handler = OrderPreservingWriterResultHandler(
                writers, compressed=compression_mode == "worker"
            )
        else:
            writer_result_handler = WriterResultHandler(
                writers, compressed=compression_mode == "worker"
            )

        self.timeout = timeout
        # Shared variable for communicating with writer thread
        self.writer_control = Control(ControlSignal.ACTIVE)
        # writer process
        self.writer_process = ResultProcess(
            writer_result_handler, result_queue, self.writer_control, timeout
        )
        self.writer_process.start()

    @property
    def is_active(self) -> bool:
        """
        Returns True if the writer process is alive and the control value is
        ControlSignal.ACTIVE.
        """
        return self.writer_process.is_alive() and self.writer_control.check_value(
            ControlSignal.ACTIVE
        )

    def set_num_batches(self, num_batches: int):
        """
        Sets the number of batches to the control variable.
        """
        self.writer_control.set_value(num_batches)

    def wait(self):
        """
        Waits for the writer process to terminate.
        """
        wait_on_process(self.writer_process, self.timeout)

    def terminate(self, retcode: ReturnCode):
        """
        Forces the writer process to terminate.
        """
        kill(self.writer_process, retcode, self.timeout)


class ParallelTrimPipelineRunner(ParallelPipelineRunner):
    """
    ParallelPipelineRunner for a TrimPipeline.
    """

    def __init__(
        self,
        command: TrimCommand,
        pipeline: TrimPipeline,
        threads: int,
        writer_manager: Optional[WriterManager] = None,
    ):
        super().__init__(command, pipeline, threads)
        self.writer_manager = writer_manager

    def ensure_alive(self):
        super().ensure_alive()
        if self.writer_manager and not self.writer_manager.is_active:
            raise MulticoreError("Writer process exited")

    def after_enqueue(self):
        # Tell the writer thread the max number of batches to expect
        if self.writer_manager:
            self.writer_manager.set_num_batches(self.num_batches)

    def finish(self):
        if self.writer_manager:
            # Wait for writer to complete
            self.writer_manager.wait()

    def terminate(self, retcode: ReturnCode):
        super().terminate(retcode)
        if self.writer_manager:
            self.writer_manager.terminate(retcode)


class QueueResultHandler(ResultHandler):
    """
    ResultHandler that writes results to the output queue.
    """

    def __init__(self, queue: Queue):
        self.queue = queue
        self.message = None
        self.timeout = None

    def start(self, worker: Optional[WorkerProcess] = None):
        if worker is None:
            raise RuntimeError("worker must not be None")

        self.message = f"{worker.name} waiting to queue result {{}}"
        self.timeout = worker.timeout

    def write_result(self, batch_num: int, result: dict):
        enqueue(
            self.queue,
            (batch_num, result),
            wait_message=self.message,
            timeout=self.timeout,
        )


class CompressingWorkerResultHandler(WorkerResultHandler):
    """
    Wraps a ResultHandler and compresses results prior to writing.
    """

    def __init__(self, *args, compression_format: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.compression_format = compression_format
        self.file_compressors: Optional[Dict[str, CompressionFormat]] = None

    def start(self, worker: Optional[WorkerProcess] = None):
        super().start(worker)
        self.file_compressors = {}

    def prepare_file(self, path: str, strings: Iterable[str]):
        compressor = self.get_compressor(path)

        if compressor:
            return (
                (path, "wb"),
                compressor.compress(b"".join(s.encode() for s in strings)),
            )
        else:
            return (path, "wt"), "".join(strings)

    def get_compressor(self, filename: str) -> CompressionFormat:
        """
        Returns the file compressor based on the file extension.
        """
        if filename not in self.file_compressors:
            self.file_compressors[filename] = xphyle.get_compressor(
                self.compression_format or filename
            )

        return self.file_compressors[filename]


class OrderPreservingWriterResultHandler(WriterResultHandler):
    """
    Writer thread that is less time/memory efficient, but is guaranteed to preserve
    the original order of records.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pending = None
        self.cur_batch = None

    def start(self, worker: Optional[WorkerProcess] = None):
        super().start(worker)
        self.pending = PendingQueue()
        self.cur_batch = 1

    def write_result(self, batch_num: int, result: dict):
        if batch_num == self.cur_batch:
            self.writers.write_result(result, self.compressed)
            self.cur_batch += 1
            self.consume_pending()
        else:
            self.pending.push(batch_num, result)

    def finish(self, total_batches: Optional[int] = None):
        if total_batches is not None:
            self.consume_pending()

            if self.cur_batch != total_batches + 1:
                raise MulticoreError(
                    f"OrderPreservingWriterResultHandler finishing "
                    f"without having seen {total_batches + 1 - self.cur_batch} of "
                    f"{total_batches} batches"
                )

        super().finish(total_batches=total_batches)

    def consume_pending(self):
        """
        Consumes any remaining items in the queue.
        """
        while (not self.pending.empty) and (
            self.cur_batch == self.pending.min_priority
        ):
            self.writers.write_result(self.pending.pop(), self.compressed)
            self.cur_batch += 1


class ResultProcess(Process):
    """
    Thread that accepts results from the worker threads and process them using a
    ResultHandler. Each batch is expected to be (batch_num, path, records),
    where path is the destination file and records is a string. Not guaranteed to
    preserve the original order of sequence records.
    """

    def __init__(
        self,
        result_handler: ResultHandler,
        queue: Queue,
        control: Control,
        timeout: int = 60,
    ):
        """
        Args:
            result_handler: A ResultHandler object.
            queue: Input queue.
            control: A shared value for communcation with the main process.
            timeout: Seconds to wait for next batch before complaining.
        """
        super().__init__(name="Result process")
        self.result_handler = result_handler
        self.queue = queue
        self.control = control
        self.timeout = timeout
        self.seen_batches = set()
        self.num_batches = None

    def run(self):
        logger.debug(f"Writer process {self.name} running under pid {os.getpid()}")

        def fail_callback():
            """
            Raises Done if the expected number of batches has been seen.
            """
            if self.num_batches is None and self.control.check_value_positive():
                self.num_batches = self.control.get_value()

            if (
                self.num_batches is not None
                and len(self.seen_batches) >= self.num_batches
            ):
                raise Done()

        def timeout_callback():
            """
            Logs an error with the missing batches.
            """
            if self.num_batches is not None:
                missing = set(range(1, self.num_batches + 1)) - self.seen_batches
                logger.error(
                    f"Result thread still missing batches"
                    f"{','.join(str(i) for i in missing)} of {self.num_batches}"
                )

        def iter_batches():
            """
            Dequeues and yields batches.
            """
            while True:
                yield dequeue(
                    self.queue,
                    wait_message="Result process waiting on result {}",
                    timeout=self.timeout,
                    fail_callback=fail_callback,
                    timeout_callback=timeout_callback,
                )

        try:
            self.result_handler.start(self)
            for batch_num, result in iter_batches():
                self.seen_batches.add(batch_num)
                self.result_handler.write_result(batch_num, result)
        except Done:
            logger.debug("Writer process exiting normally")
        except Killed:
            logger.debug("Writer process exited early")
        except:
            logger.exception("Unexpected error in writer process")
            self.control.set_value(ControlSignal.ERROR)
        finally:
            num_batches = self.control.get_value(lock=True)
            self.result_handler.finish(num_batches if num_batches > 0 else None)



import pytest
from atropos.utils import *
from atropos.utils.collections import *
from atropos.utils.ngs import *
from atropos.utils.statistics import *


def test_factorial():
    f = RandomMatchProbability()
    # simple test
    assert f.factorial(0) == 1
    assert f.factorial(1) == 1
    assert f.factorial(3) == 6
    assert int(f.factorial(27)) == int(math.factorial(27))
    # test big number
    assert int(f.factorial(150)) == int(math.factorial(150))
    # test really big number
    assert f.factorial(10000) is not None


def test_timing():
    t = Timing()
    with pytest.raises(AtroposError):
        t.summarize()

    with Timing() as t:
        time.sleep(1)
    summary = t.summarize()
    assert summary["start"] == t.start_time.isoformat
    assert summary["wallclock"] >= 1


def test_counting_dict():
    cd = CountingDict(keys=("a", "b"), sort_by=1)
    assert cd["a"] == cd["b"] == 1
    cd.increment("a")
    assert cd["a"] == 2
    assert cd["c"] == 0
    merged = cd.merge(CountingDict(keys=("b", "c")))
    assert merged["a"] == 2
    assert merged["b"] == 2
    assert merged["c"] == 1
    merged.increment("b")
    assert list(merged.summarize().keys()) == ["c", "a", "b"]
    with pytest.raises(ValueError):
        cd.merge({})


def test_histogram():
    h = Histogram(keys=(1, 1, 1, 2, 2, 3), sort_by=1)
    s = h.summarize()
    assert list(s["hist"].keys()) == [3, 2, 1]
    assert s["summary"]["mean"] == 10 / 6
    assert s["summary"]["median"] == 1.5
    assert s["summary"]["modes"] == [1]


def test_nested_dict():
    nd1 = NestedDict()
    nd1["a"].increment("x")
    nd1["a"].increment("y")

    with pytest.raises(ValueError):
        nd1.merge(1)

    nd2 = NestedDict()
    nd2["a"].increment("x")
    nd2["a"].increment("z")
    nd2["b"].increment("y")

    nd1.merge(nd2)

    assert nd1["a"]["x"] == 2
    assert nd1["a"]["y"] == 1
    assert nd1["a"]["z"] == 1
    assert nd1["b"]["y"] == 1

    assert dict(nd1.summarize()) == dict(
        columns=("x", "y", "z"), rows=dict(a=(2, 1, 1), b=(0, 1, 0))
    )

    nd1.summary_type = list
    assert set(nd1.summarize()) == {
        ("a", "x", 2),
        ("a", "y", 1),
        ("a", "z", 1),
        ("b", "y", 1),
    }


def test_merging_dict():
    obj = object()

    md = MergingDict()
    md["a"] = Const(1)
    md["b"] = dict(a=1, b=2)
    md["c"] = "foo"
    md["d"] = 5
    md["e"] = (1, 2, 3)
    md["f"] = True
    md["g"] = obj

    md2 = MergingDict()
    md2["a"] = 1
    md2["b"] = dict(a=3, b=4)
    md2["c"] = "foo"
    md2["d"] = 7
    md2["e"] = (4, 5, 6)
    md2["f"] = True
    md2["g"] = obj

    md.merge(md2)

    assert md["a"] == 1
    assert md["b"] == dict(a=4, b=6)
    assert md["c"] == "foo"
    assert md["d"] == 12
    assert md["e"] == [5, 7, 9]
    assert md["f"] is True

    failures = [
        ("a", 2, ValueError),
        ("b", "foo", TypeError),
        ("c", "bar", ValueError),
        ("f", False, ValueError),
        ("g", object(), ValueError),
    ]

    for k, v, err_type in failures:
        md3 = md2.copy()
        md3[k] = v
        with pytest.raises(err_type):
            md.merge(md3)


def test_complement():
    assert complement("ACCTGCCA") == "TGGACGGT"


def test_reverse_complement():
    assert reverse_complement("ACCTGCCA") == "TGGCAGGT"


def test_sequence_complexity():
    assert sequence_complexity("AAAA") == 0
    assert sequence_complexity("ACGT") == 2


def test_qual2int():
    assert qual2int("H") == 39


def test_quals2ints():
    assert tuple(quals2ints("HIJ")) == (39, 40, 41)


def test_qual2prob():
    assert qual2prob("+") == 0.1


def test_enumerate_range():
    assert list(enumerate_range((1, 2, 3), 5, 7)) == [(5, 1), (6, 2)]


def test_mean():
    with pytest.raises(ValueError):
        mean([])
    assert mean(range(10)) == 4.5


def test_weighted_mean():
    with pytest.raises(ValueError):
        weighted_mean([], [])
    with pytest.raises(ValueError):
        weighted_mean([1, 2], [1, 2, 3])


def test_stdev():
    with pytest.raises(ValueError):
        stdev([])
    assert stdev((1,)) == 0
    assert stdev((1, 2, 3)) == 1
    assert stdev((1, 3, 5)) == 2


def test_weighted_stdev():
    with pytest.raises(ValueError):
        weighted_stdev([], [])
    with pytest.raises(ValueError):
        weighted_stdev([1, 2], [1, 2, 3])
    assert weighted_stdev([1], [10]) == 0
    assert round(weighted_stdev([1, 2, 3], [1, 2, 3]), 4) == 0.8165


def test_median():
    with pytest.raises(ValueError):
        median([])
    assert median((1, 2, 3)) == 2
    assert median((1, 2, 3, 4)) == 2.5


def test_weighted_median():
    with pytest.raises(ValueError):
        weighted_median([], [])
    with pytest.raises(ValueError):
        weighted_median([1, 2], [1, 2, 3])
    assert weighted_median([1], [0]) is None
    assert weighted_median([1, 2, 3], [2, 2, 2]) == 2
    assert weighted_median([1, 2, 3, 4], [2, 2, 2, 2]) == 2.5
    assert weighted_median([1, 2, 3], [2, 1, 2]) == 2
    assert weighted_median([1, 2, 3], [1, 2, 3]) == 2.5


def test_modes():
    with pytest.raises(ValueError):
        modes([])
    assert modes([1]) == [1]
    assert modes([1, 2]) == [1, 2]
    assert modes([1, 1, 2]) == [1]


def test_weighted_modes():
    with pytest.raises(ValueError):
        weighted_modes([], [])
    with pytest.raises(ValueError):
        weighted_modes([1, 2], [1, 2, 3])
    assert weighted_modes([1], [1]) == [1]
    assert weighted_modes([1, 2, 3], [1, 2, 3]) == [3]
    assert weighted_modes([1, 2, 3], [1, 1, 1]) == [1, 2, 3]


def test_truncate_string():
    assert truncate_string(None) is None
    assert truncate_string("a" * 100) == ("a" * 100)
    assert truncate_string("a" * 101) == ("a" * 97) + "..."


def test_run_interruptible():
    def raise_interrupted(x=False):
        if x is True:
            raise KeyboardInterrupt()

    assert run_interruptible(raise_interrupted, False) == 0
    assert run_interruptible(raise_interrupted, True) == 130

    import errno

    def raise_io_error(x=False, y=errno.EPIPE):
        if x is True:
            err = IOError()
            err.errno = y
            raise err

    assert run_interruptible(raise_io_error, False) == 0
    assert run_interruptible(raise_io_error, True) == 1
    with pytest.raises(IOError):
        run_interruptible(raise_io_error, True, 5)

    def raises_exception(x):
        if x is True:
            raise Exception()

    assert run_interruptible(raises_exception, False) == 0
    assert run_interruptible(raises_exception, True) == 1



from enum import Enum, IntEnum
import errno
from pathlib import Path
import sys
from typing import Callable, Iterable, Iterator, Optional, Tuple, TypeVar

from loguru import logger
import pokrok
from xphyle import STDOUT, STDERR

from atropos.errors import AtroposError


T = TypeVar("T")


class LoggingConfig:
    """
    Sets up logging and retain state of whether setup has already been done.
    Specifically, don't do setup again if this function is being called externally
    such as from unit tests.
    """
    def __init__(self) -> None:
        self._was_setup = False

    @property
    def was_setup(self) -> bool:
        return self._was_setup

    def setup(
        self,
        log_path: Path,
        log_level: str,
        fmt: str = "%(asctime)s %(levelname)s: %(message)s",
        force: bool = False
    ):
        if self._was_setup and not force:
            raise ValueError("Logging was already setup")

        if log_path == STDOUT:
            sink = sys.stdout
        elif log_path == STDERR:
            sink = sys.stderr
        else:
            sink = log_path

        logger.add(sink=sink, level=log_level, format=fmt)

        self._was_setup = True


LOGGING_CONFIG = LoggingConfig()


class ReturnCode(IntEnum):
    SUCCESS = 0
    ERROR = 1
    TERMINATED = 130


class classproperty(classmethod):
    """
    Decorator for read-only class property.
    """
    def __init__(self, f):
        super().__init__(f)
        self.f = f

    def __get__(self, *args, **kwargs):
        _, owner = args
        return self.f(owner)


class Magnitude(Enum):
    """Enumeration of some common number scales.
    """

    G = ("billion", "giga", 1e9)
    M = ("million", "mega", 1e6)
    K = ("thousand", "kilo", 1e3)

    def __init__(self, label, prefix, divisor):
        self.label = label
        self.prefix = prefix
        self.divisor = divisor


def enumerate_range(
    collection: Iterable[T], start, end, step=1
) -> Iterable[Tuple[int, T]]:
    """
    Generates an indexed series:  (0,coll[0]), (1,coll[1]) ...

    Only up to (start-end+1) items will be yielded.

    Args:
        collection: The collection to enumerate.
        start: The starting index.
        end: The ending index.
        step: The amount to increment the index each iteration (defaults to 1).

    Yields:
        (index, item) tuples.
    """
    idx = start
    itr = iter(collection)
    while idx < end:
        yield idx, next(itr)
        idx += step


def truncate_string(string: Optional[str], max_len: int = 100) -> Optional[str]:
    """
    Shortens string to at most `max_len` characters, appending "..." if necessary.
    """
    if string is None:
        return None
    if len(string) > max_len:
        string = string[: max_len - 3] + "..."
    return string


def run_interruptible(func: Callable, *args, **kwargs) -> ReturnCode:
    """
    Runs a function, gracefully handling keyboard interrupts.

    Args:
        func: The function to execute.
        args, kwargs: Positional and keyword arguments to pass to `func`.

    Returns:
        A (unix-style) return code (0=normal, anything else is an error).
    """
    retcode = ReturnCode.SUCCESS
    try:
        func(*args, **kwargs)
    except KeyboardInterrupt:
        logger.error("Interrupted")
        retcode = ReturnCode.TERMINATED
    except IOError as err:
        if err.errno == errno.EPIPE:
            retcode = ReturnCode.ERROR
        else:
            raise
    except (AtroposError, EOFError):
        logger.exception("Atropos error")
        retcode = ReturnCode.ERROR
    except:  # pylint: disable=broad-except
        logger.exception("Unknown error")
        retcode = ReturnCode.ERROR
    return retcode


def create_progress_reader(
    reader: Iterable[T],
    progress_type: str = "bar",
    batch_size: int = 1,
    max_items: Optional[int] = None,
    **kwargs
) -> Iterator[T]:
    """
    Wraps an iterable in a progress bar of the specified type.

    Args:
        reader: The iterable to wrap.
        progress_type: msg = a custom progress bar that reports via log
            messages; bar = use a ProgressBar (from the progressbar library)
            or tqdm.
        max_items: Max number of items, if known in advance.
        batch_size: The number of records in each iterable item (iterable is
            typically a BatchReader).
        kwargs: Additional arguments to pass to the progress bar constructor.

    Returns:
        A wrapped iterable. If `progress_type == 'bar'` and neither of the
        supported libraries are available, a warning is logged and the unwrapped
        reader is returned.
    """
    if progress_type == "msg":
        pokrok.set_plugins(["logging"])

    if max_items:
        widgets = [
            pokrok.Widget.COUNTER,
            pokrok.Widget.PERCENT,
            pokrok.Widget.ELAPSED,
            pokrok.Widget.BAR,
            pokrok.Widget.ETA,
        ]
    else:
        widgets = [pokrok.Widget.COUNTER, pokrok.Widget.ELAPSED, pokrok.Widget.SPINNER]

    return pokrok.progress_iter(
        reader,
        desc="Processed",
        size=max_items,
        unit="records",
        multiplier=batch_size,
        style=pokrok.Style(widgets),
        **kwargs,
    )



from abc import ABCMeta, abstractmethod
import os
from pathlib import Path
from typing import (
    Callable, Iterable, Iterator, Optional, Sequence as SequenceType, Tuple, Union
)

from xphyle import xopen
from xphyle.types import ModeArg, PathOrFile
from xphyle.utils import uncompressed_size

from atropos.io import InputRead
from atropos.io.sequence import Sequence
from atropos.utils import classproperty
from atropos.utils.ngs import Alphabet
from atropos.utils.collections import Summarizable


LINESEP_LEN = len(os.linesep)


class SequenceReaderBase(
    Summarizable, Iterable[Tuple[Sequence, ...]], metaclass=ABCMeta
):
    """
    Base class for sequence readers.
    """

    @classproperty
    @abstractmethod
    def file_format(cls) -> str:
        pass

    @classproperty
    @abstractmethod
    def delivers_qualities(cls) -> bool:
        pass

    @classproperty
    @abstractmethod
    def has_qualfile(cls) -> bool:
        pass

    @classproperty
    @abstractmethod
    def colorspace(cls) -> bool:
        pass

    @classproperty
    @abstractmethod
    def interleaved(cls) -> bool:
        pass

    @classproperty
    def input_read(self) -> Optional[InputRead]:
        return None

    @property
    @abstractmethod
    def input_names(self) -> Tuple[Union[str, SequenceType[str]], Optional[str]]:
        pass

    @property
    @abstractmethod
    def quality_base(self) -> int:
        pass

    @property
    def paired(self) -> bool:
        return self.input_read == InputRead.PAIRED

    def estimate_num_records(self) -> Optional[int]:
        """
        Estimates the total number of records in the file(s), if possible.

        The way this is typically done is to divide the size of the file
        (obtained via xphyle.utils.uncompressed_size) by the size of the first
        record.

        Returns:
            An estimate of the number of records (or record pairs), or None if
            the number cannot be estimated.
        """

    @abstractmethod
    def __iter__(self) -> Iterator[Tuple[Sequence, ...]]:
        pass

    def close(self):
        pass

    def summarize(self):
        return dict(
            input_names=self.input_names,
            input_read=self.input_read,
            file_format=self.file_format,
            delivers_qualities=self.delivers_qualities,
            quality_base=self.quality_base,
            has_qualfile=self.has_qualfile,
            colorspace=self.colorspace,
            interleaved=self.interleaved,
        )


class SequenceReader(SequenceReaderBase, metaclass=ABCMeta):
    """
    Reads possibly compressed files containing sequences.
    """

    def __init__(
        self,
        path,
        mode: Optional[ModeArg] = "r",
        quality_base: Optional[int] = 33,
        alphabet: Optional[Alphabet] = None,
        sequence_factory: Callable[..., Sequence] = Sequence,
        close_on_exit: Optional[bool] = None
    ):
        """
        Args:
            path: Path or file-like object that may be compressed (.gz, .bz2, .xz).
            mode: The file open mode.
            quality_base: The minimum quality value.
            alphabet: The alphabet to use to validate sequences. If None,
                no validation is done.
            close_on_exit: Whether to close the source file on exit.
        """
        if path is None:
            raise ValueError("'path' cannot be None")

        # TODO: if quality_base is None, detect it from the data
        self._quality_base = quality_base
        self._alphabet = alphabet
        self._sequence_factory = sequence_factory

        is_path = isinstance(path, (str, Path))

        if close_on_exit is None:
            self._close_on_exit = is_path
        else:
            self._close_on_exit = close_on_exit

        if is_path:
            self.name = str(path)
            self._file = xopen(path, mode)
            if close_on_exit is None:
                self._close_on_exit = True
        else:
            if hasattr(path, "name"):
                self.name = path.name
            else:
                # TODO: generate random unique name?
                self.name = path.__class__

            self._file = path

    @property
    def input_names(self):
        return self.name, None

    @property
    def quality_base(self) -> Optional[int]:
        return self._quality_base

    def close(self):
        """
        Closes the underlying file.
        """
        if (
            self._close_on_exit and
            self._file is not None
        ):
            if hasattr(self._file, "close"):
                self._file.close()

            self._file = None

    def __enter__(self):
        if self._file is None:
            raise ValueError("I/O operation on closed SequenceReader")

        return self

    def __exit__(self, *args):
        self.close()


class PrefetchSequenceReader(SequenceReader, metaclass=ABCMeta):
    """
    SequenceReader that prefetches the first line and stores it so it can be
    inspected before iteration begins.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._seq_iter = self._iter()
        try:
            self._first_seq = next(self._seq_iter)
        except StopIteration:
            self._first_seq = None

    def __iter__(self) -> Iterator[Tuple[Sequence, ...]]:
        if self._first_seq is None:
            return
        yield self._first_seq
        yield from self._seq_iter

    @abstractmethod
    def _iter(self) -> Iterator[Tuple[Sequence, ...]]:
        pass


def estimate_num_records(
    input_file, record_size, lineseps, format_chars=0, header_size=0
) -> Optional[int]:
    """
    Convenience method to compute the estimated number of records in the input file(s).

    Args:
        input_file: Path to input file.
        record_size: The approximate size of a single record (without
            whitespace or formatting characters).
        lineseps: Number of line separators in a record (including one at
            the end).
        format_chars: Number of formatting chars, e.g. '>' in FASTA and '@'
            and '+' in FASTQ.

    Returns:
        Integer estimate of the number of records in the file.
    """
    file_size = uncompressed_size(input_file) - header_size

    if file_size is not None:
        return int(file_size / (record_size + (lineseps * LINESEP_LEN) + format_chars))


def sequence_names_match(read1: Sequence, read2: Sequence) -> bool:
    """
    Checks whether the sequences read1 and read2 have identical names, ignoring a
    suffix of '1' or '2'. Some old paired-end reads have names that end in '/1' and
    '/2'. Also, the fastq-dump tool (used for converting SRA files to FASTQ) appends
    a .1 and .2 to paired-end reads if option -I is used.

    Args:
        read1, read2: The sequences to compare.

    Returns:
        Whether the sequences are equal.
    """
    name1 = read1.name.split(None, 1)[0]
    name2 = read2.name.split(None, 1)[0]

    if name1[-1:] in "12" and name2[-1:] in "12":
        name1 = name1[:-1]
        name2 = name2[:-1]

    return name1 == name2



import os
from pathlib import Path
from typing import Optional, Tuple, Union

from xphyle import FORMATS
from xphyle.paths import as_path, check_path
from xphyle.types import ModeAccess, PathLike, PathType


def splitext_compressed(name: Union[str, Path]) -> Tuple[str, str, str]:
    """
    Splits the filename and extensions of a file that potentially has two extensions -
    one for the file type (e.g. 'fq') and one for the compression type (e.g. 'gz').

    Args:
        name: The filename.

    Returns:
        A tuple (name, ext1, ext2), where ext1 is the filetype extension and
        ext2 is the compression type extension, or None.
    """
    name_str = str(name)
    ext2 = None

    for ext in FORMATS.list_extensions(with_sep=True):
        if name_str.endswith(ext):
            ext2 = ext
            name_str = name_str[:-len(ext)]
            break

    name_str, ext1 = os.path.splitext(name_str)

    return name_str, ext1, ext2


def as_readable_path(path: Union[str, PathLike], exists: bool = True) -> Optional[Path]:
    if path is None:
        if exists:
            raise ValueError("'path' is None")
        else:
            return None

    p = as_path(path)

    if p.exists():
        check_path(p, PathType.FILE, ModeAccess.READ)
    elif exists:
        raise ValueError(f"Path {p} does not exist")

    return p



# TODO:
#  test with the --output option
#  test reading from standard input
import os

import pytest
from xphyle import open_

from atropos.console import execute_cli
from atropos.commands.trim.console import TrimCommandConsole
from atropos.utils import ReturnCode

from .utils import (
    run,
    assert_files_equal,
    datapath,
    datapathstr,
    dataurl,
    cutpath,
    redirect_stderr,
    no_internet,
    no_import,
    intercept_stdout,
    intercept_stderr,
)

try:
    import lzma
except ImportError:
    lzma = None


def test_example(tmp_path):
    run("-N -b ADAPTER", "example.fa", "example.fa", output_dir=tmp_path)


def test_example_stdout(tmp_path):
    run('-N -b ADAPTER', 'example.fa', 'example.fa', output_dir=tmp_path, stdout=True)


def test_small(tmp_path):
    run(
        "-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_empty(tmp_path):
    """empty input"""
    run(
        "-a TTAGACATATCTCCGTCG", "empty.fastq", "empty.fastq",
        output_dir=tmp_path
    )


def test_newlines(tmp_path):
    """DOS/Windows newlines"""
    run(
        "-e 0.12 -b TTAGACATATCTCCGTCG", "dos.fastq", "dos.fastq",
        output_dir=tmp_path
    )


def test_lowercase(tmp_path):
    """lowercase adapter"""
    run(
        "-b ttagacatatctccgtcg", "lowercase.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_rest(tmp_path):
    """-r/--rest-file"""
    rest_tmp = tmp_path / "rest.tmp"
    run(
        ["-b", "ADAPTER", "-N", "-r", rest_tmp], "rest.fa", "rest.fa",
        output_dir=tmp_path
    )
    assert_files_equal(datapath("rest.txt"), rest_tmp)


def test_restfront(tmp_path):
    path = tmp_path / "rest.txt"
    run(
        ["-g", "ADAPTER", "-N", "-r", path], "restfront.fa", "rest.fa",
        output_dir=tmp_path
    )
    assert_files_equal(datapath("restfront.txt"), path)


def test_discard_trimmed(tmp_path):
    """--discard"""
    run(
        "-b TTAGACATATCTCCGTCG --discard-trimmed", "discard.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_discard_untrimmed(tmp_path):
    """--discard-untrimmed"""
    run(
        "-b CAAGAT --discard-untrimmed", "discard-untrimmed.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_plus(tmp_path):
    """test if sequence name after the "+" is retained"""
    run(
        "-e 0.12 -b TTAGACATATCTCCGTCG", "plus.fastq", "plus.fastq",
        output_dir=tmp_path
    )


def test_extensiontxtgz(tmp_path):
    """automatic recognition of "_sequence.txt.gz" extension"""
    run(
        "-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz",
        output_dir=tmp_path
    )


def test_format(tmp_path):
    """the -f/--format parameter"""
    run(
        "-f fastq -b TTAGACATATCTCCGTCG", "small.fastq", "small.myownextension",
        output_dir=tmp_path

    )


def test_minimum_length(tmp_path):
    """-m/--minimum-length"""
    run(
        "-c -m 5 -a 330201030313112312", "minlen.fa", "lengths.fa",
        output_dir=tmp_path
    )


def test_too_short(tmp_path):
    """--too-short-output"""
    run(
        "-c -m 5 -a 330201030313112312 --too-short-output tooshort.tmp.fa",
        "minlen.fa",
        "lengths.fa",
        output_dir=tmp_path,
    )
    assert_files_equal(datapath("tooshort.fa"), "tooshort.tmp.fa")
    os.remove("tooshort.tmp.fa")


def test_too_short_no_primer(tmp_path):
    """--too-short-output and --trim-primer"""
    run(
        "-c -m 5 -a 330201030313112312 --trim-primer --too-short-output tooshort.tmp.fa",
        "minlen.noprimer.fa",
        "lengths.fa",
        output_dir=tmp_path,
    )
    assert_files_equal(datapath("tooshort.noprimer.fa"), "tooshort.tmp.fa")
    os.remove("tooshort.tmp.fa")


def test_maximum_length(tmp_path):
    """-M/--maximum-length"""
    run(
        "-c -M 5 -a 330201030313112312", "maxlen.fa", "lengths.fa",
        output_dir=tmp_path
    )


def test_too_long(tmp_path):
    """--too-long-output"""
    run(
        "-c -M 5 --too-long-output toolong.tmp.fa -a 330201030313112312",
        "maxlen.fa",
        "lengths.fa",
        output_dir=tmp_path,
    )
    assert_files_equal(datapath("toolong.fa"), "toolong.tmp.fa")
    os.remove("toolong.tmp.fa")


def test_length_tag(tmp_path):
    """454 data; -n and --length-tag"""
    run(
        "-n 3 -e 0.1 --length-tag length= "
        "-b TGAGACACGCAACAGGGGAAAGGCAAGGCACACAGGGGATAGG "
        "-b TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA",
        "454.fa",
        "454.fa",
        output_dir=tmp_path,
    )


def test_overlap_a(tmp_path):
    """-O/--overlap with -a (-c omitted on purpose)"""
    run(
        "-O 10 -a 330201030313112312 -e 0.0 -N", "overlapa.fa", "overlapa.fa",
        output_dir=tmp_path
    )


def test_overlap_b(tmp_path):
    """-O/--overlap with -b"""
    run(
        "-O 10 -b TTAGACATATCTCCGTCG -N", "overlapb.fa", "overlapb.fa",
        output_dir=tmp_path
    )


def test_qualtrim(tmp_path):
    """-q with low qualities"""
    run(
        "-q 10 -a XXXXXX", "lowqual.fastq", "lowqual.fastq",
        output_dir=tmp_path
    )


def test_qualbase(tmp_path):
    """-q with low qualities, using ascii(quality+64) encoding"""
    run(
        "-q 10 --quality-base 64 -a XXXXXX", "illumina64.fastq", "illumina64.fastq",
        output_dir=tmp_path
    )


def test_quality_trim_only(tmp_path):
    """only trim qualities, do not remove adapters"""
    run(
        "-q 10 --quality-base 64", "illumina64.fastq", "illumina64.fastq",
        output_dir=tmp_path
    )


def test_twoadapters(tmp_path):
    """two adapters"""
    run(
        "-a AATTTCAGGAATT -a GTTCTCTAGTTCT", "twoadapters.fasta", "twoadapters.fasta",
        output_dir=tmp_path
    )


def test_polya(tmp_path):
    """poly-A tails"""
    run(
        "-m 24 -O 10 -a AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        "polya.fasta",
        "polya.fasta",
        output_dir=tmp_path,
    )


def test_polya_brace_notation(tmp_path):
    """poly-A tails"""
    run(
        "-m 24 -O 10 -a A{35}", "polya.fasta", "polya.fasta",
        output_dir=tmp_path
    )


def test_mask_adapter(tmp_path):
    """mask adapter with N (reads maintain the same length)"""
    run(
        "-b CAAG -n 3 --mask-adapter", "anywhere_repeat.fastq", "anywhere_repeat.fastq",
        output_dir=tmp_path
    )


def test_gz_multiblock(tmp_path):
    """compressed gz file with multiple blocks (created by concatenating two .gz files)
    """
    run(
        "-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.gz",
        output_dir=tmp_path
    )


def test_suffix(tmp_path):
    """-y/--suffix parameter, combined with _F3"""
    run(
        "-c -e 0.12 -a 1=330201030313112312 -y _my_suffix_{name} --strip-f3",
        "suffix.fastq",
        "solid.csfasta",
        qualfile="solid.qual",
        output_dir=tmp_path,
    )


def test_read_wildcard(tmp_path):
    """test wildcards in reads"""
    run(
        "--match-read-wildcards -b ACGTACGT", "wildcard.fa", "wildcard.fa",
        output_dir=tmp_path
    )


def test_adapter_wildcard(tmp_path):
    """wildcards in adapter"""
    for adapter_type, expected in (
        ("-a", "wildcard_adapter.fa"),
        ("-b", "wildcard_adapter_anywhere.fa"),
    ):
        wildcardtmp = tmp_path / "wildcardtmp.txt"
        run(
            "--wildcard-file {0} {1} ACGTNNNACGT".format(wildcardtmp, adapter_type),
            expected,
            "wildcard_adapter.fa",
            output_dir=tmp_path
        )
        with open_(wildcardtmp) as wct:
            lines = wct.readlines()
        lines = [line.strip() for line in lines]
        assert lines == ["AAA 1", "GGG 2", "CCC 3b", "TTT 4b"]


def test_wildcard_n(tmp_path):
    """test 'N' wildcard matching with no allowed errors"""
    run(
        "-e 0 -a GGGGGGG --match-read-wildcards", "wildcardN.fa", "wildcardN.fa",
        output_dir=tmp_path
    )


def test_illumina_adapter_wildcard(tmp_path):
    run(
        "-a VCCGAMCYUCKHRKDCUBBCNUWNSGHCGU", "illumina.fastq", "illumina.fastq.gz",
        output_dir=tmp_path
    )


def test_adapter_front(tmp_path):
    """test adapter in front"""
    run(
        "--front ADAPTER -N", "examplefront.fa", "example.fa",
        output_dir=tmp_path
    )


def test_literal_n(tmp_path):
    """test matching literal 'N's"""
    run(
        "-N -e 0.2 -a NNNNNNNNNNNNNN", "trimN3.fasta", "trimN3.fasta",
        output_dir=tmp_path
    )


def test_literal_n2(tmp_path):
    run(
        "-N -O 1 -g NNNNNNNNNNNNNN", "trimN5.fasta", "trimN5.fasta",
        output_dir=tmp_path
    )


def test_literal_n_brace_notation(tmp_path):
    """test matching literal 'N's"""
    run(
        "-N -e 0.2 -a N{14}", "trimN3.fasta", "trimN3.fasta",
        output_dir=tmp_path
    )


def test_literal_n2_brace_notation(tmp_path):
    run(
        "-N -O 1 -g N{14}", "trimN5.fasta", "trimN5.fasta",
        output_dir=tmp_path
    )


def test_anchored_front(tmp_path):
    run(
        "-g ^FRONTADAPT -N", "anchored.fasta", "anchored.fasta",
        output_dir=tmp_path
    )


def test_anchored_front_ellipsis_notation(tmp_path):
    run(
        "-a FRONTADAPT... -N", "anchored.fasta", "anchored.fasta",
        output_dir=tmp_path
    )


def test_anchored_back(tmp_path):
    run(
        "-a BACKADAPTER$ -N", "anchored-back.fasta", "anchored-back.fasta",
        output_dir=tmp_path
    )


def test_anchored_back_no_indels(tmp_path):
    run(
        "-a BACKADAPTER$ -N --no-indels", "anchored-back.fasta", "anchored-back.fasta",
        output_dir=tmp_path
    )


def test_no_indels(tmp_path):
    run(
        "-a TTAGACATAT -g GAGATTGCCA --no-indels", "no_indels.fasta", "no_indels.fasta",
        output_dir=tmp_path
    )


def test_issue_46(tmp_path):
    """issue 46 - IndexError with --wildcard-file"""
    wildcardtmp = tmp_path / "wildcardtmp.txt"
    run(
        "--anywhere=AACGTN --wildcard-file={0}".format(wildcardtmp),
        "issue46.fasta",
        "issue46.fasta",
        output_dir=tmp_path,
    )


def test_strip_suffix(tmp_path):
    run(
        "--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta",
        output_dir=tmp_path
    )


def test_info_file(tmp_path):
    # The true adapter sequence in the illumina.fastq.gz data set is
    # GCCTAACTTCTTAGACTGCCTTAAGGACGT (fourth base is different)
    #
    infotmp = tmp_path / "infotmp.txt"
    run(
        ["--info-file", infotmp, "-a", "adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT"],
        "illumina.fastq",
        "illumina.fastq.gz",
        output_dir=tmp_path
    )
    assert_files_equal(cutpath("illumina.info.txt"), infotmp)


def test_info_file_times(tmp_path):
    infotmp = tmp_path / "infotmp.txt"
    run(
        [
            "--info-file",
            infotmp,
            "--times",
            "2",
            "-a",
            "adapt=GCCGAACTTCTTA",
            "-a",
            "adapt2=GACTGCCTTAAGGACGT",
        ],
        "illumina5.fastq",
        "illumina5.fastq",
        output_dir=tmp_path
    )
    assert_files_equal(cutpath("illumina5.info.txt"), infotmp)


def test_info_file_fasta(tmp_path):
    infotmp = tmp_path / "infotmp.txt"
    # Just make sure that it runs
    run(
        [
            "--info-file",
            infotmp,
            "-a",
            "TTAGACATAT",
            "-g",
            "GAGATTGCCA",
            "--no-indels",
        ],
        "no_indels.fasta",
        "no_indels.fasta",
        output_dir=tmp_path
    )


def test_named_adapter(tmp_path):
    run(
        "-a MY_ADAPTER=GCCGAACTTCTTAGACTGCCTTAAGGACGT",
        "illumina.fastq",
        "illumina.fastq.gz",
        output_dir=tmp_path
    )


def test_adapter_with_u(tmp_path):
    run(
        "-a GCCGAACUUCUUAGACUGCCUUAAGGACGU", "illumina.fastq", "illumina.fastq.gz",
        output_dir=tmp_path
    )


def test_no_trim(tmp_path):
    """ --no-trim """
    run(
        "--no-trim --discard-untrimmed -a CCCTAGTTAAAC", "no-trim.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_bzip2(tmp_path):
    """test bzip2 support"""
    run(
        "-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq.bz2",
        output_dir=tmp_path
    )


@pytest.mark.skipif(lzma is None, reason="no lzma library")
def test_xz(tmp_path):
    """test xz support"""
    run(
        "-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq.xz",
        output_dir=tmp_path
    )


def test_qualfile_only():
    retcode = execute_cli(["-sq", datapath("E3M.qual")])
    assert retcode == ReturnCode.ERROR


def test_no_args():
    with redirect_stderr():
        assert execute_cli() != 0


def test_anchored_no_indels(tmp_path):
    """anchored 5' adapter, mismatches only (no indels)"""
    run(
        "-g ^TTAGACATAT --no-indels -e 0.1",
        "anchored_no_indels.fasta",
        "anchored_no_indels.fasta",
        output_dir=tmp_path,
    )


def test_anchored_no_indels_wildcard_read(tmp_path):
    """anchored 5' adapter, mismatches only (no indels), but wildcards in the read
    count as matches.
    """
    run(
        "-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1",
        "anchored_no_indels_wildcard.fasta",
        "anchored_no_indels.fasta",
        output_dir=tmp_path,
    )


def test_anchored_no_indels_wildcard_adapt(tmp_path):
    """anchored 5' adapter, mismatches only (no indels), but wildcards in the adapter
    count as matches.
    """
    run(
        "-g ^TTAGACANAT --no-indels -e 0.1",
        "anchored_no_indels.fasta",
        "anchored_no_indels.fasta",
        output_dir=tmp_path,
    )


def test_unconditional_cut_front(tmp_path):
    run(
        "-u 5", "unconditional-front.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_unconditional_cut_back(tmp_path):
    run(
        "-u -5", "unconditional-back.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_unconditional_cut_both(tmp_path):
    run(
        "-u -5 -u 5", "unconditional-both.fastq", "small.fastq",
        output_dir=tmp_path
    )


def test_untrimmed_output(tmp_path):
    tmp = tmp_path / "untrimmed.tmp.fastq"
    run(
        ["-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", tmp],
        "small.trimmed.fastq",
        "small.fastq",
        output_dir=tmp_path
    )
    assert_files_equal(cutpath("small.untrimmed.fastq"), tmp)


def test_adapter_file(tmp_path):
    run(
        "-a " + dataurl("adapter.fasta"), "illumina.fastq", "illumina.fastq.gz",
        output_dir=tmp_path
    )


def test_adapter_file_5p_anchored(tmp_path):
    run(
        "-N -g " + dataurl("prefix-adapter.fasta"),
        "anchored.fasta",
        "anchored.fasta",
        output_dir=tmp_path,
    )


def test_adapter_file_3p_anchored(tmp_path):
    run(
        "-N -a " + dataurl("suffix-adapter.fasta"),
        "anchored-back.fasta",
        "anchored-back.fasta",
        output_dir=tmp_path,
    )


def test_adapter_file_5p_anchored_no_indels(tmp_path):
    run(
        "-N --no-indels -g " + dataurl("prefix-adapter.fasta"),
        "anchored.fasta",
        "anchored.fasta",
        output_dir=tmp_path,
    )


def test_adapter_file_3p_anchored_no_indels(tmp_path):
    run(
        "-N --no-indels -a " + dataurl("suffix-adapter.fasta"),
        "anchored-back.fasta",
        "anchored-back.fasta",
        output_dir=tmp_path,
    )


def test_demultiplex(tmp_path):
    multiout = str(tmp_path / "tmp-demulti.{name}.fasta")
    params = [
        "-a",
        "first=AATTTCAGGAATT",
        "-a",
        "second=GTTCTCTAGTTCT",
        "-o",
        multiout,
        "-se",
        datapathstr("twoadapters.fasta"),
    ]
    result = TrimCommandConsole.execute(params)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] == 0
    assert_files_equal(
        cutpath("twoadapters.first.fasta"), multiout.format(name="first")
    )
    assert_files_equal(
        cutpath("twoadapters.second.fasta"), multiout.format(name="second")
    )
    assert_files_equal(
        cutpath("twoadapters.unknown.fasta"), multiout.format(name="unknown")
    )


def test_max_n(tmp_path):
    run("--max-n 0", "maxn0.fasta", "maxn.fasta", output_dir=tmp_path)
    run("--max-n 1", "maxn1.fasta", "maxn.fasta", output_dir=tmp_path)
    run("--max-n 2", "maxn2.fasta", "maxn.fasta", output_dir=tmp_path)
    run("--max-n 0.2", "maxn0.2.fasta", "maxn.fasta", output_dir=tmp_path)
    run("--max-n 0.4", "maxn0.4.fasta", "maxn.fasta", output_dir=tmp_path)


def test_quiet_is_quiet():
    with intercept_stdout() as captured_stdout, intercept_stderr() as captured_stderr:
        execute_cli(
            [
                "-o",
                "/dev/null",
                "--quiet",
                "-a",
                "XXXX",
                "-se",
                datapath("illumina.fastq.gz"),
            ]
        )
        assert captured_stdout.getvalue() == ""
        assert captured_stderr.getvalue() == ""


def test_twocolor(tmp_path):
    run(
        "--twocolor-trim 22", "nextseq.fastq", "nextseq.fastq",
        output_dir=tmp_path
    )


def test_linked(tmp_path):
    run(
        "-a AAAAAAAAAA...TTTTTTTTTT", "linked.fasta", "linked.fasta",
        output_dir=tmp_path
    )


def test_fasta(tmp_path):
    run(
        "-a TTAGACATATCTCCGTCG", "small.fasta", "small.fastq",
        output_dir=tmp_path
    )


def test_custom_bisulfite_1(tmp_path):
    run(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,1", "small.fastq",
        "small.fastq", output_dir=tmp_path
    )


def test_custom_bisulfite_2(tmp_path):
    run(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 15,15,1,1",
        "small_mincut1.fastq",
        "small.fastq",
        output_dir=tmp_path,
    )


def test_custom_bisulfite_3(tmp_path):
    run(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,0",
        "small_mincut2.fastq",
        "small.fastq",
        output_dir=tmp_path,
    )


def test_custom_bisulfite_4(tmp_path):
    run(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,0,0",
        "small_mincut3.fastq",
        "small.fastq",
        output_dir=tmp_path,
    )


@pytest.mark.skipif(
    no_internet("https://ncbi.nlm.nih.gov") or no_import("srastream"),
    reason="No internet connection or srastream not importable",
)
def test_sra(tmp_path):
    run(
        "-b CTGGAGTTCAGACGTGTGCTCT --max-reads 100",
        "SRR2040662_trimmed.fq",
        sra_accn="SRR2040662",
        output_dir=tmp_path,
    )



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from enum import IntFlag
import math
import re
from typing import Iterable, Optional, Sequence as SequenceType, Tuple, Union

from loguru import logger

from atropos.adapters import AdapterCache
from atropos.aligner import Aligner, GapRule
from atropos.commands import (
    BaseCommand,
    Pipeline,
    PairedEndPipelineMixin,
    SingleEndPipelineMixin,
    Summary,
)
from atropos.io.sequence import Sequence
from atropos.utils import ReturnCode, classproperty, run_interruptible
from atropos.utils.collections import Summarizable
from atropos.utils.ngs import reverse_complement, sequence_complexity


class DetectCommand(BaseCommand):
    @classproperty
    def name(cls) -> str:
        return "detect"

    def __call__(self) -> ReturnCode:
        kmer_size = self.get_option("kmer_size", 12)
        n_reads = self.get_option("max_reads")
        overrep_cutoff = 100

        if self.get_option("include_contaminants"):
            include = Include(self.get_option("include_contaminants").upper())
        else:
            include = Include.ALL

        known_contaminants = None
        if include != Include.UNKNOWN:
            known_contaminants = self._load_known_adapters()

        detector = self.get_option("detector")
        if not detector:
            if known_contaminants and include == Include.KNOWN:
                detector = "known"
            elif n_reads <= 50000:
                detector = "heuristic"
            else:
                detector = "khmer"

        detector_args = dict(known_contaminants=known_contaminants)

        if detector == "known":
            logger.debug("Detecting contaminants using the known-only algorithm")
            detector_class = KnownContaminantDetector
            detector_args["min_kmer_match_frac"] = self.get_option(
                "min_kmer_match_frac"
            )
        elif detector == "heuristic":
            logger.debug("Detecting contaminants using the heuristic algorithm")
            detector_class = HeuristicDetector
            detector_args["min_frequency"] = self.get_option("min_frequency")
            detector_args["min_contaminant_match_frac"] = self.get_option(
                "min_contaminant_match_frac"
            )
        elif detector == "khmer":
            logger.debug("Detecting contaminants using the kmer-based algorithm")
            detector_class = KhmerDetector
        else:
            raise ValueError(f"Invalid value for 'detector': {detector}")

        summary_args = dict(
            kmer_size=kmer_size,
            n_reads=n_reads,
            overrep_cutoff=overrep_cutoff,
            include=include,
            past_end_bases=self.get_option("past_end_bases"),
        )
        detector_args.update(summary_args)

        if self.get_option("paired"):
            detector = PairedDetector(detector_class, **detector_args)
        else:
            detector = detector_class(**detector_args)

        self.summary["detect"] = summary_args
        if known_contaminants:
            self.summary["detect"][
                "known_contaminants"
            ] = known_contaminants.summarize()

        # currently only single-threaded operation is supproted
        self.summary.update(mode="serial", threads=1)

        logger.info(
            "Detecting adapters and other potential contaminant "
            "sequences based on %d-mers in %d reads",
            kmer_size,
            n_reads,
        )
        return run_interruptible(detector, self, raise_on_error=True)


class ContaminantMatcher:
    """
    Matches a known contaminant against other sequences.
    """

    @staticmethod
    def create(contaminants, kmer_size) -> SequenceType["ContaminantMatcher"]:
        """
        Creates :class:`ContaminantMatcher`s from sequences.

        Args:
            contaminants: A dict of {seq:names}.
            kmer_size: The kmer size.

        Returns:
            A list of :class:`ContaminantMatcher`s.
        """
        return [
            ContaminantMatcher(seq, names, kmer_size)
            for seq, names in contaminants.sequence_name_pairs
        ]

    def __init__(self, seq: str, names, kmer_size: int):
        """
        Args:
            seq: The contaminant sequence.
            names: Sequence of names for the contaminant.
            kmer_size: kmer size.
        """
        self.seq = seq
        self.names = names
        self.kmers = set(
            seq[i:(i + kmer_size)] for i in range(len(seq) - kmer_size + 1)
        )
        self.n_kmers = len(self.kmers)
        self.kmer_size = kmer_size
        self.matches = 0

    def match(self, seq: str, seqrc: str) -> Tuple[float, float, str]:
        """
        Matches a sequence.

        Args:
            seq: The sequence to match.
            seqrc: The reverse complement of `seq`.

        Returns:
            Tuple (f1, f2, seq), where f1 is the fraction of contaminant kmers
            that match, f2 is the fraction of sequence kmers that match, and
            seq is the best matching sequence (either `seq` or `seqrc`).
        """
        fw_kmers = set(
            seq[i:(i + self.kmer_size)] for i in range(len(seq) - self.kmer_size + 1)
        )
        fw_matches = float(len(self.kmers & fw_kmers))

        rv_kmers = set(
            seqrc[i:(i + self.kmer_size)]
            for i in range(len(seqrc) - self.kmer_size + 1)
        )
        rv_matches = float(len(self.kmers & rv_kmers))

        if fw_matches >= rv_matches:
            n_matches = fw_matches
            kmers = fw_kmers
            compare_seq = seq
        else:
            n_matches = rv_matches
            kmers = rv_kmers
            compare_seq = seqrc

        self.matches += n_matches

        match_frac1 = n_matches / self.n_kmers if self.n_kmers > 0 else 0
        match_frac2 = n_matches / len(kmers) if len(kmers) > 0 else 0

        return match_frac1, match_frac2, compare_seq


class Match(Summarizable):
    """
    A contaminant match.
    """

    def __init__(
        self,
        seq_or_contam: Union[str, ContaminantMatcher],
        count: int = 0,
        names: Optional[SequenceType[str]] = None,
        match_frac: Optional[float] = None,
        match_frac2: Optional[float] = None,
        abundance: Optional[float] = None,
        reads: Optional[Iterable[str]] = None,
    ):
        """
        Args:
            seq_or_contam: The matched sequence.
            count: The number of matches.
            names: The name(s) of matching contaminant(s).
            match_frac: The fraction of contaminant kmers that match.
            match_frac2: The fraction of sequence kmers that match.
            reads: The number of reads with the contaminant.
        """
        if isinstance(seq_or_contam, ContaminantMatcher):
            self.seq = seq_or_contam.seq
            self.count = int(seq_or_contam.matches)
            self.names = tuple(seq_or_contam.names)
            self.known_seqs = [seq_or_contam.seq]
        else:
            self.seq = seq_or_contam
            self.count = count
            self.names = tuple(names) if names else None
            self.known_seqs = None

        self.match_frac = match_frac
        self.match_frac2 = match_frac2
        self.abundance = abundance
        self.longest_match = None

        if reads:
            self.set_longest_match(reads)

    def __len__(self) -> int:
        return len(self.seq)

    def __repr__(self) -> str:
        if self.is_known:
            return f"{self.seq} => {self.names} ({self.known_seqs}))"
        else:
            return self.seq

    @property
    def seq_complexity(self) -> float:
        """
        The complexity of the sequence (0=homopolymer, 2=random).
        """
        return sequence_complexity(self.seq)

    @property
    def count_is_frequency(self) -> bool:
        """
        Whether `self.count` represents a frequency.
        """
        return isinstance(self.count, float)

    def set_contaminant(
        self,
        contam: ContaminantMatcher,
        match_frac: float,
        match_frac2: Optional[float] = None,
    ):
        """
        Set the known contaminant from a :class:`ContaminantMatcher`.
        """
        self.set_known(contam.names, [contam.seq], match_frac, match_frac2)

    def set_known(
        self,
        names: Iterable[str],
        seqs: SequenceType[str],
        match_frac: float,
        match_frac2: Optional[float] = None,
    ):
        """
        Sets the known contaminant.
        """
        self.names = tuple(names) if names else None
        self.known_seqs = seqs
        self.match_frac = match_frac
        self.match_frac2 = match_frac2

    @property
    def is_known(self) -> bool:
        """
        Whether the matching sequence is a known contaminant.
        """
        return self.known_seqs is not None

    def set_longest_match(self, sequences: Iterable[str]):
        """
        Set the longest matching sequence from a set of sequences.
        """
        for seq in sequences:
            idx = seq.index(self.seq)
            seqlen = len(self.seq) - idx
            if self.longest_match is None or self.longest_match[1] < seqlen:
                self.longest_match = (seq[idx:], seqlen)

    def estimate_abundance(self, read_sequences: Iterable[str]):
        """
        Determines whether this match's sequence is within 'seq' by simple exact
        string comparison.
        """
        self.abundance = sum(1 for read_seq in read_sequences if self.seq in read_seq)

    def summarize(self) -> dict:
        summary = dict(
            longest_kmer=self.seq,
            kmer_freq=self.count,
            kmer_freq_type="frequency" if self.count_is_frequency else "count",
            abundance=self.abundance,
            is_known=self.is_known,
            known_to_contaminant_match_frac=None,
            contaminant_to_known_match_frac=None,
            longest_match=None,
            known_names=None,
            known_seqs=None,
        )

        if self.longest_match:
            summary.update(longest_match=self.longest_match[0])

        if self.is_known:
            summary.update(
                known_to_contaminant_match_frac=self.match_frac,
                contaminant_to_known_match_frac=self.match_frac2,
                known_names=self.names,
                known_seqs=self.known_seqs,
            )

        return summary


class Include(IntFlag):
    KNOWN = 1
    UNKNOWN = 2
    ALL = KNOWN | UNKNOWN


class Detector(SingleEndPipelineMixin, Pipeline, metaclass=ABCMeta):
    """
    Base class for contaminant detectors.
    """

    def __init__(
        self,
        kmer_size: int = 12,
        n_reads: int = 10000,
        overrep_cutoff: int = 100,
        include: Include = Include.ALL,
        known_contaminants=None,
        past_end_bases: SequenceType[str] = ("A",),
    ):
        """
        Args:
            kmer_size: Size of kmers to match.
            n_reads: Number of reads to sample.
            overrep_cutoff: Degree of overrepresentation required for a kmer to be
                considered as a contaminant.
            known_contaminants: :class:`ContaminantMatcher`s to match against.
            past_end_bases: On Illumina, long runs of A (and sometimes other bases)
                can signify that the sequencer has read past the end of a fragment
                that is shorter than the read length + adapter length. Those
                bases will be removed from any sequencers before looking for
                matching contaminants.
        """
        super().__init__()

        self.kmer_size = kmer_size
        self.n_reads = n_reads
        self.overrep_cutoff = overrep_cutoff
        self.include = include
        self.known_contaminants = known_contaminants
        self._read_length = None
        self._read_sequences = set()
        self._matches = None
        self._past_end_regexp = None

        if past_end_bases:
            if len(past_end_bases[0]) > 1:
                self._past_end_regexp = re.compile(past_end_bases[0])
            else:
                self._past_end_regexp = re.compile(
                    "|".join(
                        base + "{8,}.*|" + base + "{2,}$" for base in past_end_bases
                    )
                )

    @property
    @abstractmethod
    def min_report_freq(self):
        """
        The minimum contaminant frequency required for a contaminant to be reported.
        """

    def set_read_length(self, record: Sequence):
        if self._read_length is not None:
            raise RuntimeError("Cannot set read length once it has already been set")

        self._read_length = len(record.sequence)

    def handle_records(self, context: dict, records: SequenceType[Sequence]):
        if context["size"] == 0:
            return

        if self._read_length is None:
            self.set_read_length(records[0])

        super().handle_records(context, records)

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        seq = self._filter_seq(read1.sequence)
        if seq:
            self._read_sequences.add(seq)

    def _filter_seq(self, seq: str) -> Optional[str]:
        if sequence_complexity(seq) <= 1.0:
            return None

        if self._past_end_regexp:
            match = self._past_end_regexp.search(seq)
            if match:
                seq = seq[: match.start()]

        if len(seq) < self.kmer_size:
            return None

        return seq

    def matches(self, **kwargs) -> SequenceType[Match]:
        """
        Returns the current set of matches.
        """
        if self._matches is None or len(kwargs) > 0:
            self._filter_and_sort(**kwargs)

        return self._matches

    def _filter_and_sort(
        self,
        min_len: Optional[int] = None,
        min_complexity: float = 1.1,
        min_match_frac: float = 0.1,
        limit: int = 20,
    ) -> None:
        """
        Identify, filter, and sort contaminants.

        Args:
            min_len: Minimum contaminant length.
            min_complexity: Minimum sequence complexity.
            min_match_frac: Minimum fraction of matching kmers.
            limit: Maximum number of contaminants to return.
        """
        if min_len is None:
            min_len = self.kmer_size

        matches = self._get_contaminants()
        for match in matches:
            match.estimate_abundance(self._read_sequences)

        def _filter(_match: Match):
            if _match.count < self.min_report_freq:
                logger.debug(
                    f"Filtering {_match.seq} because frequency {_match.count} < "
                    f"{self.min_report_freq}"
                )
                return False

            if min_len and len(_match) < min_len:
                logger.debug(
                    f"Filtering {_match.seq} because it's too short (< {min_len} bp)"
                )
                return False

            if min_complexity and _match.seq_complexity < min_complexity:
                logger.debug(
                    f"Filtering {_match.seq} because its complexity "
                    f"{_match.seq_complexity} < {min_complexity}"
                )
                return False

            if self.include == "known" and not _match.is_known:
                logger.debug(f"Filtering {_match.seq} because it's not known")
                return False

            elif self.include == "unknown" and _match.is_known:
                logger.debug(f"Filtering {_match.seq} because it's known")
                return False

            if (
                min_match_frac
                and _match.is_known
                and _match.match_frac < min_match_frac
            ):
                logger.debug(
                    f"Filtering {_match.seq} because its match_frac "
                    f"{_match.match_frac} < {min_match_frac}"
                )
                return False

            return True

        matches = list(
            sorted(
                filter(_filter, matches),
                key=lambda x: len(x) * math.log(x.count),
                reverse=True,
            )
        )

        if limit is not None:
            matches = matches[:limit]

        self._matches = matches

    @abstractmethod
    def _get_contaminants(self) -> Iterable[Match]:
        """
        Implemention of contaminant matching.

        Returns:
            A list of :class:`Match`es.
        """

    def finish(self, summary: Summary, **kwargs):
        super().finish(summary)
        summary["detect"]["matches"] = (
            [match.summarize() for match in self.matches(**kwargs)],
        )


class PairedDetector(PairedEndPipelineMixin, Pipeline):
    """
    Detector for paired-end reads.
    """

    def __init__(self, detector_class, **kwargs):
        super().__init__()
        self.read1_detector = detector_class(**kwargs)
        self.read2_detector = detector_class(**kwargs)
        self._read_length_set = False

    def handle_records(self, context, records):
        if context["size"] == 0:
            return

        if not self._read_length_set:
            read1, read2 = records[0]
            self.read1_detector.set_read_length(read1)
            self.read2_detector.set_read_length(read2)
            self._read_length_set = True
        super().handle_records(context, records)

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        self.read1_detector.handle_reads(context, read1)
        self.read2_detector.handle_reads(context, read2)

    def finish(self, summary, **kwargs):
        super().finish(summary)
        summary["detect"]["matches"] = (
            [match.summarize() for match in self.read1_detector.matches(**kwargs)],
            [match.summarize() for match in self.read2_detector.matches(**kwargs)],
        )


class KnownContaminantDetector(Detector):
    """
    Tests known contaminants against reads. This has linear complexity and is
    more specific than the khmer matcher, but less specific than the heuristic
    matcher. It's also less sensitive since it does not try to detect unknown
    contaminants.
    """

    def __init__(
        self,
        known_contaminants: AdapterCache,
        min_kmer_match_frac: float = 0.5,
        **kwargs
    ):
        """
        Args:
            known_contaminants: List of :class:`ContaminantMatcher`s.
            min_kmer_match_frac: Minimum fraction of matching kmers required.
            kwargs: Additional arguments to pass to the :class:`Detector`
                constructor.
        """
        super().__init__(known_contaminants=known_contaminants, **kwargs)
        self.min_kmer_match_frac = min_kmer_match_frac
        self._min_k = min(len(s) for s in known_contaminants.sequences)

    @property
    def min_report_freq(self) -> float:
        return 0.1

    def _filter_seq(self, seq: str) -> Optional[str]:
        seq = super()._filter_seq(seq)
        if seq and len(seq) >= self._min_k:
            return seq

        return None

    def _get_contaminants(self) -> Iterable[Match]:
        contaminant_matchers = ContaminantMatcher.create(
            self.known_contaminants, self.kmer_size
        )
        counts = defaultdict(int)
        max_match_fracs = defaultdict(float)

        for seq in self._read_sequences:
            seqrc = reverse_complement(seq)
            for contam in contaminant_matchers:
                match = contam.match(seq, seqrc)
                if match[0] > self.min_kmer_match_frac:
                    counts[contam] += 1
                    if match[0] > max_match_fracs[contam]:
                        max_match_fracs[contam] = match[0]

        min_count = math.ceil(
            self.n_reads
            * (self._read_length - self._min_k + 1)
            * self.overrep_cutoff
            / float(4 ** self._min_k)
        )

        return [
            Match(
                c[0],
                match_frac=max_match_fracs[c[0]],
                abundance=float(c[1]) / self.n_reads,
            )
            for c in filter(lambda x: x[1] >= min_count, counts.items())
        ]


class HeuristicDetector(Detector):
    """
    Uses a heuristic iterative algorithm to arrive at likely contaminants.
    This is the most accurate algorithm overall, but it has quadratic complexity
    and becomes too slow/memory-intenstive when n_reads > 50k.
    """

    def __init__(
        self,
        min_frequency: float = 0.001,
        min_contaminant_match_frac: float = 0.9,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.min_frequency = min_frequency
        self.min_contaminant_match_frac = min_contaminant_match_frac

    @property
    def min_report_freq(self) -> float:
        return 0.1 * self.n_reads

    def _get_contaminants(self) -> Iterable[Match]:
        def _min_count(_kmer_size: int):
            return math.ceil(
                self.n_reads
                * max(
                    self.min_frequency,
                    (
                        (self._read_length - _kmer_size + 1)
                        * self.overrep_cutoff
                        / float(4 ** _kmer_size)
                    ),
                )
            )

        kmer_size = self.kmer_size
        kmers = defaultdict(lambda: [0, set()])
        for seq in self._read_sequences:
            for i in range(len(seq) - kmer_size + 1):
                kmer = seq[i:(i + kmer_size)]
                kmers[kmer][0] += 1
                kmers[kmer][1].add(seq)

        prev = None
        cur = {}
        results = {}
        result_seqs = defaultdict(set)
        min_count = _min_count(kmer_size)

        # Identify candidate kmers for increasing values of k
        while True:
            all_seqs = set()

            for kmer, (count, seqs) in kmers.items():
                if count > min_count:
                    cur[kmer] = (count, seqs)
                    all_seqs.update(seqs)

            if len(all_seqs) == 0:
                break

            if prev:
                for kmer, (count, seqs) in prev.items():
                    if (
                        not any(seq in cur for seq in seqs)
                        and sequence_complexity(kmer) > 1.0
                    ):
                        results[kmer] = count
                        result_seqs[kmer].update(seqs)

            kmer_size += 1
            kmers = defaultdict(lambda: [0, set()])

            for seq in all_seqs:
                for i in range(len(seq) - kmer_size + 1):
                    kmer = seq[i:(i + kmer_size)]
                    kmers[kmer][0] += 1
                    kmers[kmer][1].add(seq)

            min_count = _min_count(kmer_size)
            prev = cur
            cur = {}

        results = list(results.items())

        # Now merge overlapping sequences by length and frequency to eliminate
        # redundancy in the set of candidate kmers.
        results.sort(key=lambda r: len(r[0]) * math.log(r[1]), reverse=True)
        merged = []
        unmerged = []

        while len(results) > 1:
            seq1, count1 = results[0]
            for j in range(1, len(results)):
                seq2, count2 = results[j]
                if len(seq1) >= len(seq2) and seq2 in seq1:
                    count1 += count2
                elif seq1 in seq2:
                    # if they are close in count, keep the longer sequence
                    if count1 < (2 * count2):
                        seq1 = seq2
                    count1 += count2
                else:
                    unmerged.append(results[j])

            merged.append([seq1, count1])
            results = unmerged
            unmerged = []

        results = merged + results
        if len(results) == 0:
            return []

        # TODO: For each retained match, pull out the longest sequence that
        #  matches to have a better shot of identifying long adapters that
        #  appear in full very infrequently

        # Re-sort by frequency
        results.sort(key=lambda r: r[1], reverse=True)
        # Keep anything that's within 50% of the top hit
        # TODO: make this user-configurable?
        min_count = int(results[0][1] * 0.5)
        results = (x for x in results if x[1] >= min_count)
        # Convert to matches
        matches = [Match(x[0], count=x[1], reads=result_seqs[x[0]]) for x in results]

        if self.known_contaminants:
            # Match to known sequences
            contaminants = ContaminantMatcher.create(
                self.known_contaminants, self.kmer_size
            )
            known = {}
            unknown = []

            def find_best_match(_seq, _best_matches, _best_match_frac):
                """
                Finds best contaminant matches to `seq`.
                """
                seqrc = reverse_complement(_seq)

                for _contam in contaminants:
                    match_frac1, match_frac2, compare_seq = _contam.match(_seq, seqrc)
                    if match_frac1 < _best_match_frac[0]:
                        continue

                    if _contam.seq in compare_seq or align(
                        compare_seq, _contam.seq, self.min_contaminant_match_frac
                    ):
                        if match_frac1 > _best_match_frac[0] or (
                            match_frac1 == _best_match_frac[0]
                            and match_frac2 > _best_match_frac[1]
                        ):
                            _best_matches = {}
                            _best_match_frac = (match_frac1, match_frac2)
                        _best_matches[_contam] = (match, (match_frac1, match_frac2))

                return _best_matches, _best_match_frac

            for match in matches:
                best_matches, best_match_frac = find_best_match(
                    match.seq, {}, (self.min_contaminant_match_frac, 0)
                )
                if match.longest_match:
                    best_matches, best_match_frac = find_best_match(
                        match.longest_match[0], best_matches, best_match_frac
                    )
                if best_matches:
                    for contam, _match in best_matches.items():
                        if contam not in known or _match[1] > known[contam][1]:
                            known[contam] = _match
                else:
                    unknown.append(match)

            # resolve many-many relationships
            new_matches = defaultdict(list)
            for contam, (match, match_frac) in known.items():
                new_matches[match].append((contam, match_frac))

            known = []

            for match, contams in new_matches.items():
                if len(contams) == 1:
                    contam, match_frac = contams[0]
                    match.set_contaminant(contam, *match_frac)
                else:
                    contams.sort(key=lambda x: x[1], reverse=True)
                    contam, match_frac = contams[0]
                    equiv = [
                        other_contam
                        for other_contam in contams[1:]
                        if other_contam[1] == match_frac
                    ]
                    if len(equiv) == 0:
                        match.set_contaminant(contam, *match_frac)
                    else:
                        names = {contam.names}
                        seqs = {(contam.seq,)}
                        for other_contam in equiv:
                            names.update(other_contam[0].names)
                            seqs.add(other_contam[0].seq)
                        match.set_known(list(names), list(seqs), *match_frac)
                known.append(match)

            matches = known + unknown

        return matches


class KhmerDetector(Detector):
    """
    Identifies contaminants based on kmer frequency using a fast kmer counting
    approach (as implemented in the khmer library). This approach is fast but
    not as accurate as the other two.
    """

    @property
    def min_report_freq(self) -> float:
        return 0.0001

    def _get_contaminants(self) -> Iterable[Match]:
        from khmer import Countgraph, khmer_args

        # assuming all sequences are same length
        n_win = self._read_length - self.kmer_size + 1
        tablesize = self.n_reads * n_win
        n_expected = math.ceil(tablesize / float(4 ** self.kmer_size))
        min_count = n_expected * self.overrep_cutoff
        if min_count >= 2 ** 16:
            raise ValueError(
                f"The minimum count for an over-represented k-kmer {min_count} is "
                f"greater than the max khmer count (2^16)"
            )

        countgraph = Countgraph(self.kmer_size, tablesize, khmer_args.DEFAULT_N_TABLES)
        countgraph.set_use_bigcount(True)
        for seq in self._read_sequences:
            countgraph.consume_and_tag(seq)

        candidates = {}
        for tag in countgraph.get_tagset():
            count = countgraph.get(tag)
            if count >= min_count:
                candidates[tag] = count

        if self.known_contaminants:
            matches = []
            seen = set()

            def match(_kmer):
                """Returns the frequency of `kmer` in `candidates`.
                """
                freq = candidates.get(_kmer, 0)
                if freq > 0:
                    seen.add(_kmer)
                return freq

            for seq, names in self.known_contaminants.sequence_name_pairs:
                seqlen = len(seq)
                if seqlen < self.kmer_size:
                    print(
                        f"Cannot check {list(names)[0]}; sequence is shorter than "
                        f"{self.kmer_size}"
                    )
                    continue

                n_kmers = seqlen - self.kmer_size + 1
                num_matches = 0
                match_counts = []

                for idx in range(n_kmers):
                    kmer = seq[idx:(idx + self.kmer_size)]
                    kmer_count = max(match(kmer), match(reverse_complement(kmer)))
                    if kmer_count > 0:
                        num_matches += 1
                        match_counts.append(kmer_count)

                if num_matches > 0:
                    # not sure what the correct metric is to use here
                    overall_count = sum(match_counts) / float(n_kmers)
                    matches.append(
                        Match(
                            seq,
                            count=overall_count / float(tablesize),
                            names=names,
                            match_frac=float(num_matches) / n_kmers,
                        )
                    )

            # Add remaining tags
            for tag in set(candidates.keys()) - seen:
                matches.append(Match(tag, count=candidates[tag] / float(tablesize)))
        else:
            matches = [
                Match(tag, count=count / float(tablesize))
                for tag, count in candidates.items()
            ]

        return matches


def align(seq1: str, seq2: str, min_overlap_frac: float = 0.9) -> Optional[str]:
    """
    Aligns two sequences.

    Args:
        seq1: The first sequence to align.
        seq2: The second sequence to align.
        min_overlap_frac: Minimum fraction of overlapping bases required for a
            match.

    Returns:
        The matching portion of the sequence.
    """
    aligner = Aligner(seq1, 0.0, GapRule.SEMIGLOBAL, False, False)
    aligner.min_overlap = math.ceil(min(len(seq1), len(seq2)) * min_overlap_frac)
    aligner.indel_cost = 100000
    match = aligner.locate(seq2)
    if match:
        return seq1[match[0]:match[1]]



from atropos.commands.console import (
    BaseCommandConsole, add_common_options, validate_common_options
)
from atropos.commands.detect import DetectCommand
from atropos.commands.detect.reports import DetectReportGenerator
from atropos.utils import classproperty
from atropos.utils.argparse import (
    AtroposArgumentParser,
    Namespace,
    positive,
    readable_url,
    writable_file,
    readwritable_file,
    probability,
)
from xphyle import STDOUT, STDERR


class DetectCommandConsole(DetectCommand, DetectReportGenerator, BaseCommandConsole):
    @classproperty
    def description(cls) -> str:
        return """
        Detect adapter sequences directly from read sequences. Use this command if you
        are unsure if your data has adapters, or if you know that it has adapters but
        you don't know what are the adapter sequences.
        """

    @classproperty
    def usage(cls) -> str:
        return """
        atropos detect -se input.fastq
        atropos detect -pe1 in1.fq -pe2 in2.fq
        """

    @classmethod
    def _add_arguments(cls, parser: AtroposArgumentParser) -> None:
        add_common_options(parser)
        cls._add_detect_options(parser)

    @classmethod
    def _validate_options(
        cls, options: Namespace, parser: AtroposArgumentParser
    ) -> None:
        validate_common_options(options, parser)
        cls._validate_detect_options(options, parser)

    @staticmethod
    def _add_detect_options(parser):
        parser.set_defaults(max_reads=10000, counter_magnitude="K")
        group = parser.add_group("Adapter Detection")
        group.add_argument(
            "-d",
            "--detector",
            choices=("known", "heuristic", "khmer"),
            default=None,
            help="Which detector to use. (automatically choose based on other options)",
        )
        group.add_argument(
            "-k",
            "--kmer-size",
            type=positive(),
            default=12,
            help="Size of k-mer used to scan reads for adapter sequences. (12)",
        )
        group.add_argument(
            "-e",
            "--past-end-bases",
            nargs="*",
            default=("A",),
            help="On Illumina, long runs of A (and sometimes other bases) can signify "
            "that the sequencer has read past the end of a fragment that is shorter "
            "than the read length + adapter length. Those bases will be removed from "
            "any sequencers before looking for matching contaminants. Can also be a "
            "regular expression.",
        )
        group.add_argument(
            "-i",
            "--include-contaminants",
            choices=("all", "known", "unknown"),
            default="all",
            help="What conaminants to search for: all, only known "
            "adapters/contaminants ('known'), or only unknown contaminants "
            "('unknown'). (all)",
        )
        group.add_argument(
            "-x",
            "--known-contaminant",
            action="append",
            dest="known_adapter",
            default=None,
            help="Pass known contaminants in on the commandline as "
            "'name=sequence'. Can be specified multiple times.",
        )
        group.add_argument(
            "-F",
            "--known-contaminants-file",
            type=readable_url,
            action="append",
            dest="known_adapters_file",
            default=None,
            help="Points to FASTA File or URL with known contaminants.",
        )
        group.add_argument(
            "--no-default-contaminants",
            action="store_false",
            dest="default_adapters",
            default=True,
            help="Don't fetch the default contaminant list (which is currently "
            "stored as a GitHub gist).",
        )
        group.add_argument(
            "--contaminant-cache-file",
            type=readwritable_file,
            dest="adapter_cache_file",
            default=".adapters",
            help="File where known contaminant sequences will be cached, "
            "unless --no-cache-contaminants is set.",
        )
        group.add_argument(
            "--no-cache-contaminants",
            action="store_false",
            dest="cache_adapters",
            default=True,
            help="Don't cache contaminant list as '.contaminants' in the "
            "working directory.",
        )
        group = parser.add_group("Known Detector Options")
        group.add_argument(
            "--min-kmer-match-frac",
            type=probability,
            default=0.5,
            help="Minimum fraction of contaminant kmers that must be found "
            "in a read sequence to be considered a match.",
        )
        group = parser.add_group("Heuristic Detector Options")
        group.add_argument(
            "--min-frequency",
            type=probability,
            default=0.001,
            help="Minimum frequency required to retain a k-mer.",
        )
        group.add_argument(
            "--min-contaminant-match-frac",
            type=probability,
            default=0.9,
            help="Minimum fraction of nucleotides that must align for a "
            "detected contaminant to match a known adapter sequence.",
        )
        group = parser.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writable_file,
            default=STDOUT,
            metavar="FILE",
            help="File in which to write the summary of detected adapters. (stdout)",
        )
        group.add_argument(
            "-O",
            "--output-formats",
            nargs="*",
            choices=("txt", "fasta", "json", "yaml", "pickle"),
            default=None,
            metavar="FORMAT",
            dest="report_formats",
            help="Report type(s) to generate. If multiple, '--output' is treated as a "
            "prefix and the appropriate extensions are appended. If unspecified, the "
            "format is guessed from the file extension. Supported formats are: txt, "
            "json, yaml, pickle, fasta. Additional arguments for fasta output are "
            "provided via the '--fasta' option. See the documentation for a full "
            "description of the structured output (json/yaml/pickle formats).",
        )
        group.add_argument(
            "--fasta",
            nargs="*",
            choices=("union", "perinput"),
            default=None,
            metavar="OPTIONS",
            help="Additional arguments for fasta output. Adds 'fasta' to the list of "
            "output formats if not already specified. Options are: perinput=generate "
            "one output file per input file, union=generate a single output file with "
            "all sequences merged.",
        )
        group.add_argument(
            "-m",
            "--max-adapters",
            type=positive(),
            default=None,
            help="The maximum number of candidate adapters to report. (report all)",
        )

    @staticmethod
    def _validate_detect_options(options, parser):
        options.report_file = options.output
        is_std = options.report_file in {STDOUT, STDERR}
        if options.fasta:
            if is_std and "perinput" in options.fasta:
                parser.error("Per-input fasta cannot be written to stdout")
            if not options.report_formats:
                options.report_formats = ["fasta"]
            elif "fasta" not in options.report_formats:
                options.report_formats = list(options.report_formats) + ["fasta"]
        elif is_std and options.report_formats and "fasta" in options.report_formats:
            options.fasta = ["union"]
